from .audit_log_serializer import AuditLogDefaultSerializer
from .calculation_log_serializer import CalculationLogDefaultSerializer
from .audit_log_mixin_serializer import _serialize_payload, generic_instance_payload

__all__ = [
    'AuditLogDefaultSerializer',
    'CalculationLogDefaultSerializer', 
    '_serialize_payload',
    'generic_instance_payload'
]


import os
import sys
import types
import importlib
import importlib.abc
import importlib.machinery


class ModuleAliasingFinder(importlib.abc.MetaPathFinder):
    """
    Meta path finder that:
    1. Handles both short (Folder1.Object1) and long (Project.Folder1.Object1) imports
    2. Always loads under the LONG canonical name for Django compatibility
    3. Creates aliases so both import styles work
    4. Uses ModelAwareLoader to prevent Django model re-registration
    """

    def __init__(self, project_root, repo_name):
        """
        Args:
            project_root: Base path of the project
            repo_name: Name of the repository/project (e.g., 'ArmiraCashflowDB')
        """
        self.project_root = project_root
        self.repo_name = repo_name
        self.processed_modules = set()
        self._module_aliases = {}  # Track short <-> long name mappings

    def find_spec(self, fullname, path, target=None):
        """
        Find module spec and handle aliasing between short and long names.
        Always use LONG name as canonical to ensure proper Django app_label.
        """
        # Skip modules that are part of the new app structure (lex.*)
        # These are already properly namespaced and don't need aliasing
        if fullname.startswith('lex.') and not fullname.startswith('lex_app'):
            # Check if this is a new-style app (utilities, core, authentication, etc.)
            parts = fullname.split('.')
            if len(parts) >= 2 and parts[1] in ('utilities', 'core', 'authentication', 
                                                   'audit_logging', 'api', 'process_admin'):
                return None  # Let Django handle these normally
        
        # Determine if this is a short or long name
        is_long_name = fullname.startswith(f"{self.repo_name}.")
        is_short_name = not is_long_name and self._could_be_short_name(fullname)

        if not (is_long_name or is_short_name):
            return None  # Not our module

        # Generate both short and long versions
        # ✅ ALWAYS use long name as canonical for Django compatibility
        if is_long_name:
            short_name = fullname[len(self.repo_name) + 1:]  # Remove "Project."
            long_name = fullname
        else:
            short_name = fullname
            long_name = f"{self.repo_name}.{fullname}"

        # ✅ Use LONG name as canonical
        canonical_name = long_name

        # Check if canonical (long) version is already loaded
        if canonical_name in sys.modules:
            # If requesting by short name, create alias to long name
            if fullname == short_name:
                return self._create_alias_spec(short_name, canonical_name)
            # If requesting by long name, return existing module
            return self._create_existing_spec(canonical_name)

        # Not loaded yet, find the file and load it under LONG name
        module_path = self._find_module_file(short_name)

        if module_path:
            # Import ModelAwareLoader from the same package
            from .model_loader import ModelAwareLoader
            
            # Create spec with LONG name as canonical
            is_package = module_path.endswith('__init__.py')

            spec = importlib.machinery.ModuleSpec(
                canonical_name,  # ✅ Always use long name
                ModelAwareLoader(module_path, canonical_name),
                origin=module_path,
                is_package=is_package
            )

            if is_package:
                spec.submodule_search_locations = [os.path.dirname(module_path)]

            # Store the alias mapping
            self._module_aliases[short_name] = canonical_name
            self._module_aliases[long_name] = canonical_name

            # If requesting by short name, we need to load under long name then alias
            if fullname == short_name:
                # Import the long name first
                importlib.import_module(canonical_name)
                # Now return alias spec
                return self._create_alias_spec(short_name, canonical_name)

            # Requesting by long name, return the spec directly
            return spec

        return None  # Module file not found in our project

    def _could_be_short_name(self, name):
        """Check if this could be a short name for our project."""
        # Skip if this is a new-style app (lex.*)
        if name.startswith('lex.'):
            parts = name.split('.')
            if len(parts) >= 2 and parts[1] in ('utilities', 'core', 'authentication',
                                                   'audit_logging', 'api', 'process_admin'):
                return False
        
        # Get first component
        first_part = name.split('.')[0]

        # Check if a directory with this name exists in project_root
        potential_path = os.path.join(self.project_root, first_part)

        # Only return True if it's actually a directory in our project
        if not os.path.isdir(potential_path):
            return False

        # Additional check: verify the full module path exists
        parts = name.split('.')

        # Check for package
        package_init = os.path.join(self.project_root, *parts, '__init__.py')
        if os.path.exists(package_init):
            return True

        # Check for module file
        module_file = os.path.join(self.project_root, *parts) + '.py'
        if os.path.exists(module_file):
            return True

        return False

    def _find_module_file(self, short_name):
        """
        Find the actual .py file for a module given its short name.
        Returns the full path to the file or None.
        """
        parts = short_name.split('.')

        # Try as package (__init__.py)
        package_init = os.path.join(self.project_root, *parts, '__init__.py')
        if os.path.exists(package_init):
            return package_init

        # Try as regular module (.py)
        module_file = os.path.join(self.project_root, *parts) + '.py'
        if os.path.exists(module_file):
            return module_file

        return None

    def _create_existing_spec(self, canonical_name):
        """Create a spec for an already-loaded module."""
        if canonical_name not in sys.modules:
            return None

        canonical_module = sys.modules[canonical_name]
        is_package = hasattr(canonical_module, '__path__')

        spec = importlib.machinery.ModuleSpec(
            canonical_name,
            _AliasingLoader(canonical_module, canonical_name, canonical_name),
            is_package=is_package
        )

        if is_package and hasattr(canonical_module, '__path__'):
            spec.submodule_search_locations = list(canonical_module.__path__)

        return spec

    def _create_alias_spec(self, requested_name, canonical_name):
        """
        Create a spec that aliases requested_name to canonical_name.
        """
        # Safety check: ensure canonical module exists
        if canonical_name not in sys.modules:
            return None

        canonical_module = sys.modules[canonical_name]
        is_package = hasattr(canonical_module, '__path__')

        # Create a spec that will just alias to the existing module
        spec = importlib.machinery.ModuleSpec(
            requested_name,
            _AliasingLoader(canonical_module, requested_name, canonical_name),
            is_package=is_package
        )

        if is_package and hasattr(canonical_module, '__path__'):
            spec.submodule_search_locations = list(canonical_module.__path__)

        return spec


class _AliasingLoader(importlib.abc.Loader):
    """Loader that aliases one module name to another already-loaded module."""

    def __init__(self, target_module, alias_name, canonical_name):
        self.target_module = target_module
        self.alias_name = alias_name
        self.canonical_name = canonical_name

    def create_module(self, spec):
        # Return the existing module
        return self.target_module

    def exec_module(self, module):
        # Module is already executed, just ensure aliasing in sys.modules
        if self.alias_name not in sys.modules:
            sys.modules[self.alias_name] = self.target_module

        # Also ensure parent packages exist for the alias
        self._ensure_parent_packages(self.alias_name)

    def _ensure_parent_packages(self, fullname):
        """Ensure all parent packages exist as aliases too."""
        parts = fullname.split('.')
        for i in range(1, len(parts)):
            parent_name = '.'.join(parts[:i])
            if parent_name not in sys.modules:
                # Create a minimal package module
                parent_module = types.ModuleType(parent_name)
                parent_module.__path__ = []
                sys.modules[parent_name] = parent_module

            # Set attribute on parent
            child_name = parts[i]
            child_module = sys.modules.get('.'.join(parts[:i + 1]))
            if child_module:
                setattr(sys.modules[parent_name], child_name, child_module)


from __future__ import absolute_import

import os
import logging

from celery import Celery

# set the default Django settings module for the 'celery' program.
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'lex_app.settings')

app = Celery('lex_app')

# Using a string here means the worker doesn't have to serialize
# the configuration object to child processes.
# - namespace='CELERY' means all celery-related configuration keys
#   should have a `CELERY_` prefix.
app.config_from_object('django.conf:settings', namespace='CELERY')
# app.conf.update(
#     task_serializer='dill',
#     accept_content='dill',
# )

# Load task modules from all registered Django app configs.
from django.apps import apps
app.autodiscover_tasks(lambda: [n.name for n in apps.get_app_configs()])

# Configuration validation function
def validate_celery_redis_config():
    """
    Validate that Celery is properly configured with Redis.
    This function can be called during startup to ensure configuration is correct.
    """
    logger = logging.getLogger(__name__)
    
    try:
        # Check broker connection
        broker_url = app.conf.broker_url
        result_backend = app.conf.result_backend
        
        if not broker_url.startswith('redis://'):
            logger.warning(f"Celery broker is not using Redis: {broker_url}")
            return False
            
        if not result_backend.startswith('redis://'):
            logger.warning(f"Celery result backend is not using Redis: {result_backend}")
            return False
            
        logger.info(f"Celery Redis configuration validated - Broker: {broker_url}, Backend: {result_backend}")
        return True
        
    except Exception as e:
        logger.error(f"Failed to validate Celery Redis configuration: {e}")
        return False

# Validate configuration on import (optional - can be disabled in production)
if os.getenv('CELERY_VALIDATE_CONFIG', 'True').lower() == 'true':
    validate_celery_redis_config()


"""
Context resolver for the improved CalculationLog system.

This module provides the ContextResolver class that integrates the dual context
system (operation context and model context) into a unified ContextInfo object
for use in calculation logging operations.
"""

import logging
from typing import Optional

from django.contrib.contenttypes.models import ContentType

from lex.api.utils import operation_context
from lex.audit_logging.utils.model_context import _model_context
from lex.audit_logging.utils.data_models import ContextInfo, ContextResolutionError

logger = logging.getLogger(__name__)


class ContextResolver:
    """
    Resolves and integrates dual context systems for calculation logging.
    
    This class combines operation context (from operation_context) and model context
    (from ModelContext) into a unified ContextInfo object that can be used
    throughout the calculation logging system.
    """
    
    @staticmethod
    def resolve() -> ContextInfo:
        """
        Resolve current context by integrating both context systems.
        
        Extracts:
        - Operation context from operation_context (calculation_id, request_obj)
        - Model context from ModelContext (current/parent models)
        
        Returns:
            ContextInfo: Unified context information for logging operations
            
        Raises:
            ContextResolutionError: When required context information is missing
                or when AuditLog cannot be resolved
        """
        try:
            # Extract calculation_id from operation context
            context_data = operation_context.get()
            calculation_id = context_data.get('calculation_id')
            
            if not calculation_id:
                raise ContextResolutionError(
                    "Missing calculation_id in operation context",
                    calculation_id=calculation_id
                )
            
            # Resolve AuditLog using calculation_id
            # Import here to avoid circular dependency
            from lex.audit_logging.models.audit_log import AuditLog
            
            try:
                audit_log = context_data.get('audit_log_temp') or AuditLog.objects.get(calculation_id=calculation_id)
                if audit_log.calculation_id == None:
                    audit_log.calculation_id = calculation_id
                    audit_log.save()

            except AuditLog.DoesNotExist:
                raise ContextResolutionError(
                    f"AuditLog not found for calculation_id: {calculation_id}",
                    calculation_id=calculation_id
                )
            except Exception as e:
                raise ContextResolutionError(
                    f"Error retrieving AuditLog for calculation_id {calculation_id}: {str(e)}",
                    calculation_id=calculation_id
                )
            
            # Extract model context using new ModelContext structure
            model_ctx = _model_context.get()['model_context']
            
            # Determine current and parent models using direct property access
            current_model = model_ctx.current
            parent_model = model_ctx.parent
            current_record = None
            parent_record = None
            content_type = None
            parent_content_type = None
            root = model_ctx.get_root()
            root_record = None
            
            # Process current model if it exists
            if current_model:
                try:
                    content_type = ContentType.objects.get_for_model(current_model)
                    current_record = f"{current_model._meta.model_name}_{current_model.pk}"
                except Exception as e:
                    logger.warning(
                        f"Error resolving ContentType for current model: {e}",
                        extra={'calculation_id': calculation_id}
                    )
            if root:
                try:
                    root_record = f"{root._meta.model_name}_{root.pk}"
                except Exception as e:
                    logger.warning(
                        f"Error resolving ContentType for root model: {e}",
                        extra={'calculation_id': calculation_id}
                    )

            # Process parent model if it exists
            if parent_model:
                try:
                    parent_content_type = ContentType.objects.get_for_model(parent_model)
                    parent_record = f"{parent_model._meta.model_name}_{parent_model.pk}"
                except Exception as e:
                    logger.warning(
                        f"Error resolving ContentType for parent model: {e}",
                        extra={'calculation_id': calculation_id}
                    )
            
            # Create and return unified ContextInfo
            context_info = ContextInfo(
                calculation_id=calculation_id,
                audit_log=audit_log,
                current_model=current_model,
                parent_model=parent_model,
                current_record=current_record,
                parent_record=parent_record,
                content_type=content_type,
                parent_content_type=parent_content_type,
                root_record=root_record
            )
            
            logger.debug(
                f"Context resolved successfully for calculation_id: {calculation_id}",
                extra={
                    'calculation_id': calculation_id,
                    'has_current_model': current_model is not None,
                    'has_parent_model': parent_model is not None
                }
            )
            
            return context_info
            
        except ContextResolutionError:
            # Re-raise ContextResolutionError as-is
            raise
        except Exception as e:
            # Wrap any other exceptions in ContextResolutionError
            calculation_id = None
            try:
                calculation_id = operation_context.get('calculation_id')
            except Exception:
                pass
            
            raise ContextResolutionError(
                f"Unexpected error during context resolution: {str(e)}",
                calculation_id=calculation_id,
                has_current_model=model_ctx.current,
                has_parent_model=model_ctx.parent
            )


class FilterTreeNode:
    def __init__(self, node_id, model_container, parent_to_self_fk_name, selection, children):
        self.node_id = node_id
        self.model_container = model_container
        self.parent_to_self_fk_name = parent_to_self_fk_name
        self.selection = selection
        self.children = children

        # QuerySet of objects filtered at this node
        self.filtered_objects = {}

    # Finds all currently filtered objects at this node and writes them into 'self.filtered_objects'
    def evaluate(self):
        # In case the current node has no more children, set 'self.filtered_objects' to
        #  'model.objects.all()'
        if not self.children:
            self.filtered_objects = self.model_container.model_class.objects.all()
            return

        child_to_selected_filtered_objects_dict = {}
        for child in self.children:
            child.evaluate()
            # Fill 'child_to_selected_filtered_objects_dict[child]' with
            #  'child.filtered_objects' intersected with the child's selection
            child_to_selected_filtered_objects_dict[
                child] = child.filtered_objects if child.selection == 'noSelection' else child.filtered_objects.filter(
                **{
                    self.model_container.pk_name + '__in': child.selection
                })

        # Aggregate all selected filtered objects for each child and filter accordingly
        filters = {
            child.parent_to_self_fk_name + '__in': child_to_selected_filtered_objects_dict[child]
            for child
            in self.children
        }
        self.filtered_objects = self.model_container.model_class.objects.filter(**filters)

    # Writes for this node an entry of type 'self.node_id -> list(pk) at that id filtered at that node'
    # into the passed dictionary
    def write_self_to_dict(self, d):
        d[self.node_id] = list(map(
            lambda obj: obj.pk,
            list(set(self.filtered_objects))
        ))

        for child in self.children:
            child.write_self_to_dict(d)



from lex.core.signals.calculation_signals import (
    update_calculation_status,
    do_post_save,
    custom_post_save,
    create_profile,
    save_profile
)

__all__ = [
    'update_calculation_status',
    'do_post_save',
    'custom_post_save',
    'create_profile',
    'save_profile'
]



# core/management/commands/Init.py
import json
import time
import uuid
import logging
import traceback
from copy import deepcopy
from pathlib import Path

import requests
from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.apps import apps
from django.db import connection
from django.db.migrations.executor import MigrationExecutor
from django.db.migrations.questioner import MigrationQuestioner
from django.db.migrations.state import ProjectState
from django.db.migrations.loader import MigrationLoader
from django.db.migrations.autodetector import MigrationAutodetector
from django.db.migrations.operations.models import CreateModel, DeleteModel, RenameModel
from typing import Dict, List, Tuple, Set, Any

from core.management.commands.bootstrap_callback_server import start_callback_server

logger = logging.getLogger(__name__)
# core/management/commands/Init.py

import os
import uuid
import webbrowser
from urllib.parse import urlencode

from django.conf import settings
import os
import time
import uuid
import json
import socket
import asyncio
import threading
import webbrowser
from pathlib import Path
from urllib.parse import urlencode

import uvicorn
from django.conf import settings
import os


KEYCLOAK_ENV_VARS = [
    "KEYCLOAK_URL",
    "KEYCLOAK_REALM",
    "KEYCLOAK_REALM_NAME",
    "OIDC_RP_CLIENT_ID",
    "OIDC_RP_CLIENT_SECRET",
    "OIDC_RP_CLIENT_UUID",
]

STATE_FILE = Path(
    getattr(settings, "KEYCLOAK_STATE_FILE", Path(os.getcwd() or settings.BASE_DIR) / ".keycloak_state.json")
)
ENV_FILE = Path(os.getcwd() or settings.BASE_DIR) / ".env"


def get_missing_keycloak_env() -> list[str]:
    missing = []
    for key in KEYCLOAK_ENV_VARS:
        if key in ("KEYCLOAK_REALM", "KEYCLOAK_REALM_NAME"):
            if not (
                os.getenv("KEYCLOAK_REALM")
                or os.getenv("KEYCLOAK_REALM_NAME")
                or getattr(settings, "KEYCLOAK_REALM_NAME", None)
            ):
                missing.append("KEYCLOAK_REALM/KEYCLOAK_REALM_NAME")
            continue
        if not (os.getenv(key) or getattr(settings, key, None)):
            missing.append(key)
    return missing


def build_instance_controller_url(state: str, callback_url: str) -> str:
    base = getattr(settings, "INSTANCE_CONTROLLER_BASE_URL", "").rstrip("/")
    if not base:
        raise RuntimeError("INSTANCE_CONTROLLER_BASE_URL is not configured")
    params = {
        "state": state,
        "callback": callback_url,
        "flow": "keycloak-client-bootstrap",
        "project": getattr(settings, "LEX_PROJECT_SLUG", "lex-app"),
    }
    return f"{base}/lex/keycloak-bootstrap?{urlencode(params)}"


def _load_state_map() -> dict:
    if not STATE_FILE.exists():
        return {}
    try:
        return json.loads(STATE_FILE.read_text(encoding="utf-8") or "{}")
    except Exception:
        return {}


def get_state(state: str) -> str | None:
    data = _load_state_map()
    return data.get(state)


def wait_for_keycloak_setup(state: str, timeout_seconds: int = 900, poll_interval: int = 3) -> bool:
    start = time.time()
    while True:
        if not get_missing_keycloak_env():
            print("Keycloak credentials detected; continuing Init.")
            return True

        current = get_state(state)
        if current == "done":
            print("Keycloak setup completed; continuing Init.")
            return True
        if current == "cancelled":
            print("Keycloak setup cancelled; stopping Init.")
            return False

        if time.time() - start > timeout_seconds:
            print(f"Timed out waiting for Keycloak setup after {timeout_seconds} seconds.")
            return False

        time.sleep(poll_interval)

def _port_in_use(host: str, port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.settimeout(0.5)
        return s.connect_ex((host, port)) == 0


def start_local_server_if_needed(host: str = "127.0.0.1", port: int = 9002) -> None:
    if _port_in_use(host, port):
        return

    def run_uvicorn():
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        uvicorn.run(
            "lex_app.asgi:application",
            host=host,
            port=port,
            loop="asyncio",
            reload=False,
            log_level="info",
        )

    t = threading.Thread(target=run_uvicorn, daemon=True)
    t.start()



def _save_state_map(data: dict) -> None:
    tmp = STATE_FILE.with_suffix(".tmp")
    tmp.write_text(json.dumps(data), encoding="utf-8")
    tmp.replace(STATE_FILE)


def set_state(state: str, value: str) -> None:
    data = _load_state_map()
    data[state] = value
    _save_state_map(data)




def poll_bootstrap_status(state: str, timeout_seconds: int = 900, poll_interval: int = 3) -> dict | None:
    base = getattr(settings, "INSTANCE_CONTROLLER_BASE_URL", "").rstrip("/")
    if not base:
        raise RuntimeError("INSTANCE_CONTROLLER_BASE_URL not configured")

    url = f"{base}/api/lex/bootstrap_status/{state}"
    start = time.time()

    while True:
        if time.time() - start > timeout_seconds:
            print(f"Timed out waiting for Keycloak setup after {timeout_seconds} seconds.")
            return None

        try:
            resp = requests.get(url, timeout=10)
            resp.raise_for_status()
            data = resp.json()
        except Exception as e:
            print(f"Error polling bootstrap status: {e}")
            time.sleep(poll_interval)
            continue

        status_val = data.get("status")
        if status_val == "done":
            payload = data.get("payload")
            if payload:
                print("Keycloak credentials retrieved from instance-controller.")
                return payload
            else:
                print("Status is done but payload is missing; retrying...")
                time.sleep(poll_interval)
                continue
        elif status_val == "cancelled":
            print("Keycloak setup cancelled.")
            return None

        # still pending, keep polling
        time.sleep(poll_interval)


def update_env_file(values: dict) -> None:
    env_file = Path(settings.BASE_DIR) / ".env"
    text = ""
    if env_file.exists():
        text = env_file.read_text(encoding="utf-8")

    lines = text.splitlines()
    kv = {}
    non_kv_lines = []

    for line in lines:
        stripped = line.strip()
        if not stripped or stripped.startswith("#") or "=" not in stripped:
            non_kv_lines.append(line)
            continue
        k, v = stripped.split("=", 1)
        kv[k] = v

    for k, v in values.items():
        kv[k] = v

    out_lines = []
    out_lines.extend(non_kv_lines)
    for k, v in kv.items():
        out_lines.append(f"{k}={v}")

    tmp = env_file.with_suffix(".env.tmp")
    tmp.write_text("\n".join(out_lines) + "\n", encoding="utf-8")
    tmp.replace(env_file)

def initiate_device_flow():
    state = str(uuid.uuid4())
    # Your equivalent of /oauth/device/authorize
    base = getattr(settings, "INSTANCE_CONTROLLER_BASE_URL", "").rstrip("/")
    resp = requests.post(
        f"{base}/api/lex/bootstrap/initiate",
        json={"state": state},
    )
    resp.raise_for_status()
    data = resp.json()
    # returns: {"state": "...", "verification_uri": "https://.../lex/keycloak-bootstrap?state=..."}
    return data["state"], data["verification_uri"]

def poll_for_credentials(state, interval=5, timeout=900):
    start = time.time()
    while True:
        if time.time() - start > timeout:
            return None  # timeout

        base = getattr(settings, "INSTANCE_CONTROLLER_BASE_URL", "").rstrip("/")
        resp = requests.get(
            f"{base}/api/lex/bootstrap_status/{state}"
        )
        resp.raise_for_status()
        data = resp.json()

        if data.get("status") == "done":
            return data["payload"]  # dict with 5 env vars
        if data.get("status") == "cancelled":
            return None

        time.sleep(interval)

def fetch_bootstrap_state(state: str) -> dict | None:
    base = getattr(settings, "INSTANCE_CONTROLLER_BASE_URL", "").rstrip("/")

    url = f"{base}/api/lex/bootstrap_status/{state}"
    resp = requests.get(url, timeout=5)
    resp.raise_for_status()
    data = resp.json()
    if data.get("status") == "done" and data.get("payload"):
        return data["payload"]
    if data.get("status") == "cancelled":
        return {"cancelled": True}
    return None

def wait_for_keycloak_setup_v2(state: str, timeout_seconds: int = 900, poll_interval: int = 3) -> dict | None:
    start = time.time()
    while True:
        # if env already set, short-circuit
        if not get_missing_keycloak_env():
            return {
                "KEYCLOAK_URL": os.getenv("KEYCLOAK_URL", ""),
                "KEYCLOAK_REALM": os.getenv("KEYCLOAK_REALM", ""),
                "OIDC_RP_CLIENT_ID": os.getenv("OIDC_RP_CLIENT_ID", ""),
                "OIDC_RP_CLIENT_SECRET": os.getenv("OIDC_RP_CLIENT_SECRET", ""),
                "OIDC_RP_CLIENT_UUID": os.getenv("OIDC_RP_CLIENT_UUID", ""),
            }

        state_data = fetch_bootstrap_state(state)
        if state_data:
            if state_data.get("cancelled"):
                return None
            return state_data  # dict with 5 keys

        if time.time() - start > timeout_seconds:
            return None

        time.sleep(poll_interval)

class KeycloakSyncManager:
    """Handles the sync between Django models and Keycloak resources/permissions"""

    def __init__(self):
        from lex.api.views.authentication.KeycloakManager import KeycloakManager
        # Parse .env file and update os.environ
        from dotenv import dotenv_values
        env_config = dotenv_values('.env')
        for key, value in env_config.items():
            os.environ[key] = value
        self.kc_manager = KeycloakManager()
        self.default_scopes = ['list', 'read', 'create', 'edit', 'delete', 'export']
        self.exported_configs = None

    def get_all_django_models(self) -> Set[str]:
        """Get all Django model names in the format 'app_label.ModelName'"""
        all_models = set()

        for app_config in apps.get_app_configs():
            app_label = app_config.label

            # Skip Django's built-in apps that we don't want to sync
            skip_apps = {
                'admin', 'auth', 'contenttypes', 'sessions', 'messages',
                'staticfiles', 'migrations', 'django_extensions'
            }

            if app_label in skip_apps:
                logger.debug(f"Skipping built-in app: {app_label}")
                continue

            try:
                for model in app_config.get_models():
                    # Skip abstract models
                    if model._meta.abstract:
                        logger.debug(f"Skipping abstract model: {app_label}.{model.__name__}")
                        continue

                    # Skip proxy models (they don't need separate permissions)
                    if model._meta.proxy:
                        logger.debug(f"Skipping proxy model: {app_label}.{model.__name__}")
                        continue

                    model_name = f"{app_label}.{model.__name__}"
                    all_models.add(model_name)
                    logger.debug(f"Found model: {model_name}")

            except Exception as e:
                logger.warning(f"Error getting models for app {app_label}: {e}")

        logger.info(f"Found {len(all_models)} Django models total")
        return all_models


    def export_configs(self):
        if self.exported_configs:
            return self.exported_configs
        self.exported_configs = self.kc_manager.export_authorization_settings()
        return self.exported_configs

    def get_existing_keycloak_resources(self, auth_config: Dict) -> Set[str]:
        """Get all resource names that exist in Keycloak config"""
        existing_resources = set()

        for resource in auth_config.get('resources', []):
            resource_name = resource.get('name')
            if resource_name:
                existing_resources.add(resource_name)

        logger.info(f"Found {len(existing_resources)} existing Keycloak resources")
        return existing_resources

    def find_missing_models(self, all_django_models: Set[str], existing_keycloak_resources: Set[str],
                            to_delete_set: Set[str]) -> Set[str]:
        """Find Django models that are missing from Keycloak and should be added"""
        # Models that exist in Django but not in Keycloak (excluding ones we're about to delete)
        missing_models = all_django_models - existing_keycloak_resources - to_delete_set

        if missing_models:
            logger.info(f"Found {len(missing_models)} models missing from Keycloak:")
            for model_name in sorted(missing_models):
                logger.info(f"  Missing: {model_name}")
        else:
            logger.info("No missing models found - all Django models are synced with Keycloak")

        return missing_models
    def find_permissions_for_resource_name(self, resource_name: str, auth_config: Dict) -> List[Dict[str, Any]]:
        """Find all permissions that reference a specific resource name in the policies array"""
        resource_permissions = []

        for policy in auth_config.get('policies', []):
            if policy.get('type') == 'scope':  # This is actually a permission
                config = policy.get('config', {})
                resources_str = config.get('resources', '[]')
                try:
                    resources = json.loads(resources_str) if isinstance(resources_str, str) else resources_str
                    if resource_name in resources:
                        resource_permissions.append(policy)
                except:
                    pass

        return resource_permissions

    def delete_resources_individual(self, to_delete_set: Set[str], all_resources: List[Dict]) -> bool:
        """Delete resources using individual API calls - reuse existing resource data"""
        try:
            deletion_success = 0

            logger.info(f"Using existing resource data with {len(all_resources)} resources")

            # Create a mapping using the resource data we already have (NO additional API calls!)
            resource_name_to_id = {}

            for resource in all_resources:
                resource_name = resource.get('name')
                if resource_name in to_delete_set:
                    # Get the _id field from the existing resource data
                    resource_id = resource.get('_id')
                    if resource_id:
                        resource_name_to_id[resource_name] = resource_id
                        logger.info(f"Found resource to delete: {resource_name} -> {resource_id}")

            logger.info(f"Found {len(resource_name_to_id)} resources to delete")

            # Delete each resource found
            for resource_name in to_delete_set:
                if resource_name not in resource_name_to_id:
                    logger.warning(f"Resource {resource_name} not found, may already be deleted")
                    deletion_success += 1
                    continue

                resource_id = resource_name_to_id[resource_name]

                try:
                    # Get permissions for this resource and delete them first
                    permissions = self.kc_manager.admin.get_client_authz_permissions(
                        client_id=self.kc_manager.client_uuid
                    )

                    for permission in permissions:
                        perm_resources = permission.get('resources', [])
                        if resource_id in perm_resources:
                            perm_id = permission.get('id')
                            perm_name = permission.get('name')

                            if perm_id:
                                try:
                                    self.kc_manager.admin.delete_client_authz_permission(
                                        client_id=self.kc_manager.client_uuid,
                                        permission_id=perm_id
                                    )
                                    logger.info(f"    ✓ Deleted permission: {perm_name}")
                                except Exception as e:
                                    logger.error(f"    ✗ Failed to delete permission {perm_name}: {e}")
                                    return False

                    # Delete the resource itself - use UMA API for resource deletion
                    try:
                        self.kc_manager.uma.resource_set_delete(resource_id)
                        logger.info(f"  ✓ Deleted resource: {resource_name}")
                        deletion_success += 1
                    except Exception as e:
                        logger.error(f"  ✗ Failed to delete resource {resource_name}: {e}")
                        return False

                except Exception as e:
                    logger.error(f"Error deleting resource {resource_name}: {e}")
                    return False

            logger.info(f"Successfully deleted {deletion_success}/{len(to_delete_set)} resources")
            return deletion_success == len(to_delete_set)

        except Exception as e:
            logger.error(f"Error in bulk resource deletion: {e}")
            return False

    def ensure_default_authz(self, auth_config: Dict, ensure_default_authz: bool) -> None:
        if not ensure_default_authz:
            return

        resources = auth_config.setdefault('resources', [])
        policies = auth_config.setdefault('policies', [])

        # 1) Default Resource (represents all resources)
        default_resource_name = "Default Resource"
        has_default_resource = any(r.get('name') == default_resource_name for r in resources)
        if not has_default_resource:
            resources.append({
                "name": default_resource_name,
                "type": "urn:keycloak:resource:default",
                "ownerManagedAccess": False,
                "attributes": {},
                # Represent *all* application URLs
                "uris": ["/*"],
                "scopes": []  # no specific scopes needed for the global default
            })

        # 2) Regex-based Default Policy (instead of JS)
        default_policy_name = "Default Policy"
        has_default_policy = any(
            p.get('name') == default_policy_name and p.get('type') == 'regex'
            for p in policies
        )
        if not has_default_policy:
            policies.append({
                "name": default_policy_name,
                "type": "regex",
                "logic": "POSITIVE",
                "decisionStrategy": "UNANIMOUS",
                # In export-style JSON, provider-specific config usually lives in 'config'
                "config": {
                    # Always-match regex on a stable claim, e.g. preferred_username
                    "pattern": ".*",
                    "targetClaim": "preferred_username"
                }
            })

        # 3) Resource-based Default Permission
        default_permission_name = "Default Permission"
        has_default_permission = any(
            p.get('name') == default_permission_name and p.get('type') == 'resource'
            for p in policies
        )
        if not has_default_permission:
            policies.append({
                "name": default_permission_name,
                "type": "resource",  # resource-based permission
                "logic": "POSITIVE",
                "decisionStrategy": "AFFIRMATIVE",
                "config": {
                    # Resource-based: tie it to the Default Resource by *name*
                    "resources": f'["{default_resource_name}"]',
                    # No scopes for pure resource permission
                    "applyPolicies": f'["{default_policy_name}"]'
                }
            })

    def ensure_core_role_policies(self, auth_config: Dict) -> None:
        """
        Ensures the three core role-based policies exist in the auth_config.
        These are required for all model permissions to work.
        """
        policies = auth_config.setdefault('policies', [])

        # Define the three core role-based policies
        core_roles = ['admin', 'standard', 'view-only']

        # Check which policies already exist
        existing_policy_names = {p.get('name') for p in policies}

        for role_name in core_roles:
            full_policy_name = f"Policy - {role_name}"

            if full_policy_name not in existing_policy_names:
                # Get the role ID from Keycloak (roles must exist beforehand)
                try:
                    role = self.kc_manager.admin.get_client_role(
                        client_id=self.kc_manager.client_uuid,
                        role_name=role_name
                    )
                    role_id = role.get('id')

                    if not role_id:
                        logger.warning(f"Role '{role_name}' not found - creating it first")
                        # Create the role if it doesn't exist
                        role_payload = {
                            "name": role_name,
                            "description": f"Role for {role_name} policy",
                            "clientRole": True,
                            "composite": False,
                        }
                        self.kc_manager.admin.create_client_role(
                            client_role_id=self.kc_manager.client_uuid,
                            payload=role_payload,
                            skip_exists=True,
                        )
                        # Fetch the role again to get its ID
                        role = self.kc_manager.admin.get_client_role(
                            client_id=self.kc_manager.client_uuid,
                            role_name=role_name
                        )
                        role_id = role.get('id')

                    # Create the role-based policy in the config
                    policy = {
                        "name": full_policy_name,
                        "type": "role",
                        "logic": "POSITIVE",
                        "decisionStrategy": "UNANIMOUS",
                        "config": {
                            "roles": json.dumps([{"id": role_id, "required": True}])
                        }
                    }
                    policies.append(policy)
                    logger.info(f"   ✓ Added core policy to config: {full_policy_name}")

                except Exception as e:
                    logger.error(f"Failed to ensure policy {full_policy_name}: {e}")
            else:
                logger.info(f"   ✓ Core policy already exists: {full_policy_name}")

    def process_model_changes(self, adds: List[Tuple[str, str]], deletes: List[Tuple[str, str]], 
                            renames: List[Tuple[str, str, str]], preserve_permissions: bool = True,     ensure_default_authz: bool = False,) -> bool:
        """Process all model changes using export-operate-delete-import strategy"""

        try:
            # 1. Export current authorization settings (single export)
            logger.info("Exporting current authorization settings...")
            auth_config = self.export_configs()
            if not auth_config:
                logger.error("Failed to export authorization settings")
                return False

            # 2. Get all resources with complete data (single API call)
            logger.info("Getting all resources with complete data...")
            all_resources = self.kc_manager.admin.get_client_authz_resources(
                client_id=self.kc_manager.client_uuid
            )
            logger.info(f"Retrieved {len(all_resources)} resources with complete data")

            # Ensure required sections exist in auth_config
            if 'resources' not in auth_config:
                auth_config['resources'] = []
            if 'policies' not in auth_config:
                auth_config['policies'] = []

            logger.info(f"Auth config structure: {list(auth_config.keys())}")
            logger.info(f"Found {len(auth_config.get('resources', []))} resources in config")
            logger.info(f"Found {len(auth_config.get('policies', []))} policies in config")

            # 3. Build sets for processing
            to_delete_set = set()
            to_add_set = set()
            preserved_permissions = {}

            # Process deletions
            for app_label, model_name in deletes:
                resource_name = f"{app_label}.{model_name}"
                to_delete_set.add(resource_name)

            # Process renames - save permissions and add to delete set
            for app_label, old_name, new_name in renames:
                old_resource_name = f"{app_label}.{old_name}"
                new_resource_name = f"{app_label}.{new_name}"

                to_delete_set.add(old_resource_name)
                to_add_set.add(new_resource_name)

                if preserve_permissions:
                    # Find the old resource in the exported config
                    old_resource = None
                    for resource in auth_config['resources']:
                        if resource.get('name') == old_resource_name:
                            old_resource = resource
                            break

                    if old_resource:
                        old_permissions = self.find_permissions_for_resource_name(old_resource_name, auth_config)

                        preserved_permissions[new_resource_name] = {
                            'resource': old_resource,
                            'permissions': old_permissions
                        }
                        logger.info(f"  ✓ Preserved {len(old_permissions)} permissions for {old_resource_name} -> {new_resource_name}")
                    else:
                        logger.warning(f"  ⚠ Resource {old_resource_name} not found in config for preservation")

            # Process additions
            for app_label, model_name in adds:
                resource_name = f"{app_label}.{model_name}"
                to_add_set.add(resource_name)

            # 4. Delete resources from exported config (operate on exported data)
            resources_to_keep = []
            policies_to_keep = []


            to_delete_set_config = deepcopy(to_delete_set)
            # to_delete_set_config.add("Default Resource")


            for resource in auth_config['resources']:
                resource_name = resource.get('name')
                if resource_name not in to_delete_set_config:
                    resources_to_keep.append(resource)
                else:
                    logger.info(f"  ✓ Removed {resource_name} from config")

            # Remove permissions that reference deleted resources (permissions are in policies array)
            for policy in auth_config['policies']:
                if (True or
                        policy.get('name', '').strip() not in ["Default Policy", "Default Permission"]):
                    if policy.get('type') == 'scope':  # This is a permission
                        config = policy.get('config', {})
                        resources_str = config.get('resources', '[]')
                        try:
                            resources = json.loads(resources_str) if isinstance(resources_str, str) else resources_str
                            # Check if any referenced resource is being deleted
                            if not any(res_name in to_delete_set_config for res_name in resources):
                                policies_to_keep.append(policy)
                            else:
                                logger.info(f"  ✓ Removed permission {policy.get('name')} from config")
                        except:
                            policies_to_keep.append(policy)  # Keep if we can't parse
                    else:
                        # Regular policy, keep it
                        policies_to_keep.append(policy)

            auth_config['resources'] = resources_to_keep
            auth_config['policies'] = policies_to_keep

            # 5. Send individual deletion requests (use existing resource data - NO extra API calls)
            if to_delete_set:
                logger.info(f"Deleting {len(to_delete_set)} resources individually...")
                if not self.delete_resources_individual(to_delete_set, all_resources):
                    logger.error("Failed to delete resources individually")
                    return False

            # 6. Add new resources and renamed resources to config
            if to_add_set:
                logger.info(f"Adding {len(to_add_set)} resources to config...")

                # Find policy names for standard permissions (no IDs needed)
                available_policy_names = set()
                for policy in auth_config.get('policies', []):
                    policy_name = policy.get('name')
                    if policy_name in ['Policy - admin', 'Policy - standard', 'Policy - view-only']:
                        available_policy_names.add(policy_name)

                logger.info(f"Found {len(available_policy_names)} policy names: {list(available_policy_names)}")

                # Define scope-policy mapping using policy NAMES (not IDs)
                scope_policy_mapping = {
                    'list': ['Policy - admin', 'Policy - standard', 'Policy - view-only'],
                    'read': ['Policy - admin', 'Policy - standard', 'Policy - view-only'],
                    'create': ['Policy - admin'],
                    'edit': ['Policy - admin', 'Policy - standard'],
                    'delete': ['Policy - admin'],
                    'export': ['Policy - admin', 'Policy - standard']
                }
                permissions_to_create = []

                for resource_name in to_add_set:
                    try:
                        # Check if this is a renamed model with preserved permissions
                        if resource_name in preserved_permissions:
                            # Use preserved data
                            preserved_data = preserved_permissions[resource_name]
                            old_resource = preserved_data['resource']
                            old_permissions = preserved_data['permissions']

                            # Create new resource with preserved attributes but new name (NO _id field)
                            new_resource = {
                                "name": resource_name,
                                "type": old_resource.get("type", "django-model"),
                                "ownerManagedAccess": old_resource.get("ownerManagedAccess", False),
                                "attributes": old_resource.get("attributes", {}),
                                "uris": old_resource.get("uris", []),
                                "scopes": old_resource.get("scopes", [{"name": scope} for scope in self.default_scopes])
                            }
                            auth_config['resources'].append(new_resource)

                            # Recreate permissions with new resource name in policies array
                            for old_permission in old_permissions:
                                # Update permission name to reflect new resource name
                                old_perm_name = old_permission.get('name', '')
                                old_resource_name = preserved_data['resource']['name']
                                new_perm_name = old_perm_name.replace(old_resource_name, resource_name)

                                # Convert to proper Keycloak permission format (in policies array)
                                old_config = old_permission.get('config', {})
                                new_permission = {
                                    "name": new_perm_name,
                                    "type": "scope",
                                    "logic": old_permission.get("logic", "POSITIVE"),
                                    "decisionStrategy": old_permission.get("decisionStrategy", "AFFIRMATIVE"),
                                    "config": {
                                        "resources": f'["{resource_name}"]',  # Use new resource name
                                        "scopes": old_config.get("scopes", "[]"),
                                        "applyPolicies": old_config.get("applyPolicies", "[]")
                                    }
                                }
                                auth_config['policies'].append(new_permission)

                            logger.info(f"  ✓ Added renamed resource {resource_name} with preserved permissions")

                        else:
                            # Create new resource with default permissions
                            # Check if resource already exists
                            existing_resource = None
                            for resource in auth_config['resources']:
                                if resource.get('name') == resource_name:
                                    existing_resource = resource
                                    break

                            if existing_resource:
                                logger.info(f"  ✓ Resource {resource_name} already exists in config")
                                continue

                            # Create new resource (NO _id field)
                            new_resource = {
                                "name": resource_name,
                                "ownerManagedAccess": False,
                                "attributes": {},
                                "uris": [],
                                "scopes": [{"name": scope} for scope in self.default_scopes]
                            }
                            auth_config['resources'].append(new_resource)

                            # Create default permissions for each scope (add to policies array)
                            for scope, policy_names in scope_policy_mapping.items():
                                applicable_policies = [name for name in policy_names if name in available_policy_names]
                                if applicable_policies:
                                    new_permission = {
                                        "name": f"Permission - {resource_name} - {scope}",
                                        "type": "scope",
                                        "logic": "POSITIVE",
                                        "decisionStrategy": "AFFIRMATIVE",
                                        "config": {
                                            "resources": json.dumps([resource_name]),  # Resource by name
                                            "scopes": json.dumps([scope]),  # Scope by name
                                            "applyPolicies": json.dumps(applicable_policies)  # Policy by name
                                        }
                                    }
                                    auth_config['policies'].append(new_permission)
                                    logger.info(
                                        f"   ✓ Added permission for scope '{scope}' with policies: {applicable_policies}")
                                else:
                                    logger.warning(f"No applicable policies found for scope {scope}")

                            logger.info(f"  ✓ Added new resource {resource_name} with default permissions")

                    except Exception as e:
                        logger.error(f"Error adding resource {resource_name}: {e}")
                        logger.error(f"Full traceback: {traceback.format_exc()}")
                        return False

            # After step 6 (Add new resources) and before step 7 (Import)
            # ...existing code that adds resources...

            # 6.5. Ensure core role-based policies exist
            logger.info("Ensuring core role-based policies exist...")
            self.ensure_core_role_policies(auth_config)

            # Also ensure default authz if requested
            if ensure_default_authz:
                logger.info("Ensuring default authorization settings...")
                self.ensure_default_authz(auth_config, ensure_default_authz)

            # Log summary before import
            logger.info(f"Total resources to import: {len(auth_config['resources'])}")
            logger.info(f"Total policies to import: {len(auth_config['policies'])}")

            # 7. Import the updated configuration (single import)
            logger.info("Importing updated authorization settings...")
            success = self.kc_manager.import_authorization_settings(auth_config)

            # 7. Import the updated configuration (single import)
            logger.info("Importing updated authorization settings...")
            success = self.kc_manager.import_authorization_settings(auth_config)

            if success:
                logger.info("Successfully imported updated authorization settings")
                return True
            else:
                logger.error("Failed to import updated authorization settings")
                return False

        except Exception as e:
            logger.error(f"Error processing model changes: {e}")
            logger.error(f"Full traceback: {traceback.format_exc()}")
            return False


class Command(BaseCommand):
    help = "Sync Keycloak authorization settings with Django model changes"

    def add_arguments(self, parser):
        parser.add_argument(
            '--dry-run',
            action='store_true',
            help='Show what would be changed without making actual changes',
        )
        parser.add_argument(
            '--preserve-renamed-permissions',
            action='store_true',
            default=True,
            help='Preserve permissions when renaming models (default: True)',
        )
        parser.add_argument(
            '--check-missing',
            action='store_true',
            default=True,
            help='Check for Django models missing from Keycloak and add them (default: True)',
        )
        parser.add_argument(
            '--ensure-default-authz',
            action='store_true',
            default=False,
            help='Ensure a default resource, regex default policy and resource-based default permission exist'
        )

    def check_unapplied_migrations(self):
        """Check if there are unapplied migrations (migration files that exist but haven't been applied to DB)"""
        from django.db.migrations.executor import MigrationExecutor
        from django.db import connection

        executor = MigrationExecutor(connection)
        plan = executor.migration_plan(executor.loader.graph.leaf_nodes())
        return len(plan) > 0

    def execute_migrations(self, verbosity=1, create_new=True):
        """Execute Django migrations with proper workflow"""
        try:
            success = True

            # Step 1: Check for model changes and create new migrations if needed
            if create_new:
                self.stdout.write("Creating new migrations for model changes...")
                call_command(
                    'makemigrations',
                    verbosity=verbosity,
                    interactive=False,
                    stdout=self.stdout,
                    stderr=self.stderr,
                    no_input=True
                )
                self.stdout.write("✓ New migrations created successfully")

            # Step 2: Check for unapplied migrations
            if not self.check_unapplied_migrations():
                self.stdout.write("No unapplied migrations found.")
                return success

            # Step 3: Apply unapplied migrations
            self.stdout.write("Applying unapplied migrations...")
            call_command(
                'migrate',
                verbosity=verbosity,
                interactive=False,
                stdout=self.stdout,
                stderr=self.stderr
            )
            
            
            # Execute migrations
            call_command(
                'createcachetable',
                verbosity=verbosity,
            )

            self.stdout.write("✓ Django migrations completed successfully")
            return success

        except Exception as e:
            self.stderr.write(f"✗ Migration failed: {e}")
            logger.error(f"Migration execution failed: {e}")
            logger.error(f"Full traceback: {traceback.format_exc()}")
            return False
        

    def handle(self, *args, **options):
        tour = options.get("tour", False)
        dry_run = options.get('dry_run', False)
        preserve_permissions = options.get('preserve_renamed_permissions', True)
        check_missing = options.get('check_missing', True)
        skip_migrations = options.get('skip_migrations', False)
        migration_verbosity = options.get('migration_verbosity', 1)
        ensure_default_authz = options.get('ensure_default_authz', False)

        missing = get_missing_keycloak_env()
        if missing:
            state = str(uuid.uuid4())

            # Start minimal callback server on ephemeral port
            server, port = start_callback_server(
                state=state,
                env_file=ENV_FILE,
                state_file=STATE_FILE,
                host="127.0.0.1",
                port=0,  # auto-pick
            )

            callback_url = f"http://127.0.0.1:{port}/callback"

            try:
                url = build_instance_controller_url(state, callback_url)
            except Exception as e:
                self.stderr.write(f"Missing Keycloak env: {', '.join(missing)}")
                self.stderr.write(f"Cannot build instance-controller URL: {e}")
                return

            self.stdout.write("")
            self.stdout.write("Keycloak credentials are not configured.")
            self.stdout.write(f"Missing: {', '.join(missing)}")
            self.stdout.write("")
            self.stdout.write(f"Local callback server started on 127.0.0.1:{port}")
            self.stdout.write("Open this URL in your browser to complete setup:")
            self.stdout.write(f"  {url}")
            self.stdout.write(f"State token: {state}")
            self.stdout.write("Waiting for setup to complete (Ctrl+C to abort)...")
            self.stdout.write("")

            try:
                webbrowser.open(url)
            except Exception:
                pass

            ok = wait_for_keycloak_setup(state, timeout_seconds=900, poll_interval=3)
            if not ok:
                self.stderr.write("Init aborted; Keycloak not configured.")
                return

            self.stdout.write("Keycloak configured. Continuing Init...")

        self.stdout.write("=" * 80)
        self.stdout.write("Django Migration + Keycloak Authorization Sync")
        self.stdout.write("=" * 80)
        # Step 1: Initialize sync manager
        try:
            sync_manager = KeycloakSyncManager()
        except Exception as e:
            self.stderr.write(f"Failed to initialize Keycloak manager: {e}")
            return

        # Step 2: CRITICAL
        self.stdout.write("Detecting model changes BEFORE migrations...")

        # Detect model changes
        questioner = MigrationQuestioner(defaults={"ask_rename_model": True})
        loader = MigrationLoader(None, ignore_no_migrations=True)
        autodetector = MigrationAutodetector(
            loader.project_state(),
            ProjectState.from_apps(apps),
            questioner=questioner,
        )
        changes = autodetector.changes(graph=loader.graph)

        # Process changes
        adds, deletes, renames = [], [], []
        missing_models = {}

        for app_label, migrations in changes.items():
            for migration in migrations:
                for operation in migration.operations:
                    if isinstance(operation, CreateModel):
                        adds.append((app_label, operation.name))
                    elif isinstance(operation, DeleteModel):
                        deletes.append((app_label, operation.name))
                    elif isinstance(operation, RenameModel):
                        renames.append((app_label, operation.old_name, operation.new_name))

        # Display detected changes
        if adds or deletes or renames:
            self.stdout.write("Detected model changes:")

            for app, name in adds:
                self.stdout.write(f"  ADD {app}.{name}")

            for app, name in deletes:
                self.stdout.write(f"  DELETE {app}.{name}")

            for app, old, new in renames:
                self.stdout.write(f"  RENAME {app}.{old} -> {new}")
        else:
            self.stdout.write("No model changes detected.")
            # return

        if not skip_migrations:
            self.stdout.write("\n" + "-" * 80)
            self.stdout.write("Executing Django Migrations")
            self.stdout.write("-" * 80)

            if dry_run:
                self.stdout.write("DRY RUN: Would execute Django migrations")
                if self.check_unapplied_migrations():
                    self.stdout.write("Pending migrations found - would be executed")
                else:
                    self.stdout.write("No pending migrations found")
            else:
                if not self.execute_migrations(migration_verbosity):
                    self.stderr.write("Migration failed - aborting Keycloak sync")
                    return
        else:
            self.stdout.write("\nSkipping Django migrations (--skip-migrations)")


        if dry_run:
            self.stdout.write("Dry run mode - no changes will be made.")



        if check_missing:
            auth_config = sync_manager.export_configs()

            if auth_config:
                all_django_models = sync_manager.get_all_django_models()
                existing_keycloak_resources = sync_manager.get_existing_keycloak_resources(auth_config)
                missing_models = sync_manager.find_missing_models(all_django_models, existing_keycloak_resources, set())

                for app_name, old_name, new_name in renames:
                    if f"{app_name}.{new_name}" in missing_models:
                        missing_models.remove(f"{app_name}.{new_name}")

                if missing_models:
                    self.stdout.write(f"Would add {len(missing_models)} missing models:")
                    for model_name in sorted(missing_models):
                        self.stdout.write(f"  WOULD ADD: {model_name}")
                else:
                    self.stdout.write("No missing models found.")

        if dry_run:
            return

        # Process all changes using the optimized strategy
        self.stdout.write("\nSyncing changes to Keycloak...")
        missing_models = {tuple(model.split(".")) for model in missing_models}
        missing_models.union(set(adds))
        new_missing_models_list = list(missing_models)

        success = sync_manager.process_model_changes(new_missing_models_list, deletes, renames, preserve_permissions, ensure_default_authz=ensure_default_authz,
)

        if success:
            self.stdout.write("✓ All model changes successfully synced to Keycloak!")
        else:
            self.stdout.write("✗ Some operations failed. Check logs for details.")


from lex.process_admin.models.model_collection import ModelCollection
from lex.process_admin.models.model_container import ModelContainer
from lex.process_admin.models.model_process_admin import ModelProcessAdmin

__all__ = ['ModelCollection', 'ModelContainer', 'ModelProcessAdmin']



# Core mixins
# Import mixins explicitly when needed to avoid circular dependencies
# during Django initialization

__all__ = [
    'CalculatedModelMixin', 
    'ModelCombinationGenerator', 
    'ModelClusterManager', 
    'calc_and_save_sync',
    'AdminReportsModificationRestriction', 
    'ExampleModelModificationRestriction'
]


import json

from rest_framework import filters


# TODO: test this
class UserReadRestrictionFilterBackend(filters.BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        model_container = view.kwargs['model_container']
        return self._filter_queryset(request, queryset, model_container)

    def _filter_queryset(self, request, queryset, model_container):
        model = model_container.model_class
        if hasattr(model, 'modification_restriction'):
            pks = [obj.pk for obj in queryset if model.modification_restriction.can_be_read(instance=obj, user=request.user, violations=[])]
            queryset.filter(pk__in=pks)
        return queryset


def create_filter_queries_from_tree_paths(all_filter_queries, filter_node, query_string_so_far):
    if 'entries' in filter_node:
        all_filter_queries[query_string_so_far + 'in'] = filter_node['entries']
    else:
        for key, value in filter_node['children'].items():
            new_query_string = query_string_so_far + key + '__'
            create_filter_queries_from_tree_paths(all_filter_queries, value, new_query_string)


class ForeignKeyFilterBackend(filters.BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        active_filter_tree = request.GET.get('activeFilterTree', None)
        all_filter_queries = {}
        if active_filter_tree is not None:
            active_filter_tree = json.loads(active_filter_tree)
            create_filter_queries_from_tree_paths(all_filter_queries, active_filter_tree, '')
        return queryset.filter(**all_filter_queries)


class PrimaryKeyListFilterBackend(filters.BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        model_container = view.kwargs['model_container']
        filter_arguments = {
            model_container.pk_name + '__in':
                list(filter(lambda x: x != '', request.query_params.dict()['pks'].split(',')))
        } if 'pks' in request.query_params.dict() else {}
        return queryset.filter(**filter_arguments)


class StringFilterBackend(filters.BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        filter_arguments = request.GET.get('searchParams', '{}')
        return queryset.filter(**json.loads(filter_arguments))


"""
Django management command to check which models have or will have history tracking.
"""

from django.core.management.base import BaseCommand
from django.apps import apps
from django.db import models
from lex.lex_app.simple_history_config import should_track_model, get_model_exclusion_reason


class Command(BaseCommand):
    help = 'Check which models have or will have history tracking enabled'

    def add_arguments(self, parser):
        parser.add_argument(
            '--app',
            type=str,
            help='Check models only from specific app',
        )
        parser.add_argument(
            '--show-excluded',
            action='store_true',
            help='Show models that are excluded from history tracking',
        )
        parser.add_argument(
            '--show-tracked',
            action='store_true',
            help='Show models that have history tracking',
        )
        parser.add_argument(
            '--show-all',
            action='store_true',
            help='Show all models (tracked and excluded)',
        )

    def handle(self, *args, **options):
        app_label = options.get('app')
        show_excluded = options.get('show_excluded', False)
        show_tracked = options.get('show_tracked', False)
        show_all = options.get('show_all', False)
        
        # If no specific flags are set, show all by default
        if not (show_excluded or show_tracked):
            show_all = True

        tracked_models = []
        excluded_models = []
        already_tracked_models = []

        # Get all models
        if app_label:
            try:
                app_config = apps.get_app_config(app_label)
                models_to_check = app_config.get_models()
            except LookupError:
                self.stdout.write(
                    self.style.ERROR(f'App "{app_label}" not found.')
                )
                return
        else:
            models_to_check = apps.get_models()

        # Categorize models
        for model in models_to_check:
            if hasattr(model, 'history'):
                already_tracked_models.append(model)
            elif should_track_model(model):
                tracked_models.append(model)
            else:
                excluded_models.append(model)

        # Display results
        self.stdout.write(
            self.style.SUCCESS(f'\n=== History Tracking Status ===\n')
        )

        if show_all or show_tracked:
            self.stdout.write(
                self.style.SUCCESS(f'✓ Models WITH history tracking ({len(already_tracked_models)}):')
            )
            for model in sorted(already_tracked_models, key=lambda m: f"{m._meta.app_label}.{m.__name__}"):
                self.stdout.write(f'  • {model._meta.app_label}.{model.__name__}')
            
            self.stdout.write(
                self.style.WARNING(f'\n→ Models that WILL GET history tracking ({len(tracked_models)}):')
            )
            for model in sorted(tracked_models, key=lambda m: f"{m._meta.app_label}.{m.__name__}"):
                self.stdout.write(f'  • {model._meta.app_label}.{model.__name__}')

        if show_all or show_excluded:
            self.stdout.write(
                self.style.ERROR(f'\n⊘ Models EXCLUDED from history tracking ({len(excluded_models)}):')
            )
            for model in sorted(excluded_models, key=lambda m: f"{m._meta.app_label}.{m.__name__}"):
                reason = get_model_exclusion_reason(model)
                self.stdout.write(f'  • {model._meta.app_label}.{model.__name__} - {reason}')

        # Summary
        total_models = len(tracked_models) + len(excluded_models) + len(already_tracked_models)
        self.stdout.write(
            self.style.SUCCESS(f'\n=== Summary ===')
        )
        self.stdout.write(f'Total models: {total_models}')
        self.stdout.write(f'Already tracked: {len(already_tracked_models)}')
        self.stdout.write(f'Will be tracked: {len(tracked_models)}')
        self.stdout.write(f'Excluded: {len(excluded_models)}')
        
        if tracked_models:
            self.stdout.write(
                self.style.WARNING(f'\nTo apply history tracking to new models, run: python manage.py makemigrations')
            )


from django.http import JsonResponse

from django.views import View
class HealthCheck(View):
    authentication_classes = []

    def get(self, request):
        return JsonResponse({"status": "Healthy :)"})



import traceback
import logging

from django.db import transaction
from rest_framework_api_key.permissions import HasAPIKey

from lex.audit_logging.utils.model_context import model_logging_context
from rest_framework.exceptions import APIException
from rest_framework.generics import RetrieveUpdateDestroyAPIView, CreateAPIView
from rest_framework.mixins import CreateModelMixin, UpdateModelMixin

from rest_framework.response import Response
from rest_framework import status
from lex.core.models.calculation_model import CalculationModel

from lex.core.signals import update_calculation_status
from lex.audit_logging.mixins.audit_mixin import AuditLogMixin
from lex.api.utils.context import OperationContext
from lex.api.views.model_entries.mixins.DestroyOneWithPayloadMixin import (
    DestroyOneWithPayloadMixin,
)
from lex.api.views.model_entries.mixins.ModelEntryProviderMixin import (
    ModelEntryProviderMixin,
)
from rest_framework.permissions import IsAuthenticated
from rest_framework.exceptions import PermissionDenied

from lex.api.views.permissions.UserPermission import UserPermission
from lex.audit_logging.utils.cache_manager import CacheManager
from lex.audit_logging.utils.websocket_notifier import WebSocketNotifier

logger = logging.getLogger(__name__)


class OneModelEntry(
    AuditLogMixin,
    ModelEntryProviderMixin,
    DestroyOneWithPayloadMixin,
    RetrieveUpdateDestroyAPIView,
    CreateAPIView,
):
    permission_classes = [IsAuthenticated]

    def create(self, request, *args, **kwargs):
        model_container = self.kwargs["model_container"]
        instance = model_container.model_class()
        
        # Check create permission using new system
        try:
            if hasattr(instance, 'permission_create'):
                from lex.core.models.base import UserContext
                user_context = UserContext.from_request(request, instance)
                can_create = instance.permission_create(user_context)
            else:
                # Fallback to legacy method
                can_create = instance.can_create(request)
                
            if not can_create:
                return Response(
                    {
                        "message": f"You are not authorized to create a record in {model_container.model_class.__name__}"
                    },
                    status=status.HTTP_400_BAD_REQUEST
                )
        except Exception:
            # Allow by default on permission check error
            pass

            # return Response(data={}, status=status.HTTP_204_NO_CONTENT, headers={}, exception=e)

        calculationId = self.kwargs["calculationId"]

        with OperationContext(request, calculationId) as context_id:

            try:
                with transaction.atomic():
                    response = CreateModelMixin.create(self, request, *args, **kwargs)

            except Exception as e:
                raise APIException(
                    {"error": f"{e} ", "traceback": traceback.format_exc()}
                )

            return response

    def update(self, request, *args, **kwargs):

        model_container = self.kwargs["model_container"]
        calculationId = self.kwargs["calculationId"]
        instance = model_container.model_class()

        with OperationContext(request, calculationId):
            instance = model_container.model_class.objects.filter(
                pk=self.kwargs["pk"]
            ).first()
            with model_logging_context(instance):
                if "calculate" in request.data and request.data["calculate"] == "true":
                    # instance = model_container.model_class.objects.filter(pk=self.kwargs["pk"]).first()
                    instance.untrack()
                    instance.is_calculated = CalculationModel.IN_PROGRESS
                    instance.save(skip_hooks=True)
                    calculation_id = calculationId
                    calculation_record = f"{instance._meta.model_name}_{instance.pk}"
                    WebSocketNotifier.send_calculation_update(
                        calculation_id=calculationId,
                        calculation_record=f"{instance._meta.model_name}_{instance.pk}"
                    )
                    cache_key = CacheManager.build_cache_key(
                        calculation_record,
                        calculation_id
                    )
                    CacheManager.store_message(cache_key, "")
                    instance.track()
                    update_calculation_status(instance)

                # TODO: For sharepoint preview, find a new way to create an audit log with the new structure
                # if "edited_file" not in request.data:

                try:
                    response = UpdateModelMixin.update(self, request, *args, **kwargs)


                except Exception as e:
                    raise APIException(
                        {"error": f"{e} ", "traceback": traceback.format_exc()}
                    )

                return response



import base64
import json
from typing import Optional, Dict
import os
import traceback
import urllib.parse
import jwt
import requests
from datetime import datetime, timezone
import logging

import streamlit as st

from lex.api.views.authentication.KeycloakManager import KeycloakManager
from django.conf import settings

logger = logging.getLogger(__name__)


import threading
import time
import requests
import os
import jwt
import logging
from streamlit.runtime.scriptrunner import add_script_run_ctx, get_script_run_ctx

log = logging.getLogger(__name__)

# -------------------------
# Token refresh: config
# -------------------------
TOKEN_SKEW_SECONDS = 10          # refresh 60s before exp
REFRESH_MIN_INTERVAL = 15        # floor sleep
REFRESH_MAX_BACKOFF = 300        # cap backoff to 5 minutes

def _oidc_token_endpoint() -> str:
    base = os.getenv("KEYCLOAK_URL").rstrip("/")
    realm = os.getenv("KEYCLOAK_REALM")
    return f"{base}/realms/{realm}/protocol/openid-connect/token"

def _decode_exp_no_verify(token: str) -> int:
    try:
        claims = jwt.decode(token, options={"verify_signature": False, "verify_exp": False})
        return int(claims.get("exp", 0)) if claims else 0
    except Exception:
        return 0

def _now() -> int:
    return int(time.time())

def _compute_next_refresh_at(exp: int, expires_in: int | None) -> int:
    if exp:
        return max(_now() + REFRESH_MIN_INTERVAL, exp - TOKEN_SKEW_SECONDS)
    if expires_in:
        return _now() + max(REFRESH_MIN_INTERVAL, int(expires_in) - TOKEN_SKEW_SECONDS)
    # Unknown expiry: retry soon
    return _now() + REFRESH_MIN_INTERVAL

def _post_refresh(refresh_token: str) -> dict | None:
    url = _oidc_token_endpoint()
    # client_id = os.getenv("KEYCLOAK_CLIENT_ID") or ""
    # client_secret = os.getenv("KEYCLOAK_CLIENT_SECRET") or None
    client_id = os.getenv("OIDC_RP_CLIENT_ID")
    client_secret = os.getenv("OIDC_RP_CLIENT_SECRET")
    data = {
        "grant_type": "refresh_token",
        "refresh_token": refresh_token,
        "client_id": client_id,
    }
    if client_secret:
        data["client_secret"] = client_secret
    try:
        r = requests.post(url, data=data, timeout=15)
        if r.status_code >= 400:
            log.warning("Refresh failed: %s %s", r.status_code, r.text)
            return None
        return r.json()
    except Exception as e:
        log.warning("Refresh exception: %s", e)
        return None

def _update_tokens_from_response(tok: dict) -> None:
    access = tok.get("access_token") or ""
    refresh = tok.get("refresh_token") or st.session_state.get("refresh_token") or ""
    expires_in = tok.get("expires_in")
    exp = _decode_exp_no_verify(access) if access else 0
    st.session_state.access_token = access
    st.session_state.refresh_token = refresh
    st.session_state.token_exp = exp
    st.session_state.expires_in = expires_in
    # Optionally hydrate user info again if needed by downstream code
    # user = get_user_info(access) or st.session_state.get("user_info")

def _token_refresher(stop_key: str = "stop_token_refresher") -> None:
    backoff = 5
    while not st.session_state.get(stop_key, False):
        access = st.session_state.get("access_token") or ""
        refresh = st.session_state.get("refresh_token") or ""
        exp = st.session_state.get("token_exp") or _decode_exp_no_verify(access)
        expires_in = st.session_state.get("expires_in")

        # Decide next refresh time
        next_at = _compute_next_refresh_at(exp, expires_in)
        sleep_for = max(1, next_at - _now())
        # Sleep with small increments so stop flag is responsive
        end_at = _now() + sleep_for
        while _now() < end_at:
            if st.session_state.get(stop_key, False):
                return
            time.sleep(min(1.0, end_at - _now()))

        if st.session_state.get(stop_key, False):
            return

        if not refresh:
            # Nothing to refresh; try again later
            backoff = min(REFRESH_MAX_BACKOFF, backoff * 2)
            time.sleep(backoff)
            continue

        tok = _post_refresh(refresh)
        print("Access Token", tok.get('access_token'))
        if tok and tok.get("access_token"):
            _update_tokens_from_response(tok)
            backoff = 5  # reset backoff on success
        else:
            # Failure: backoff and retry
            backoff = min(REFRESH_MAX_BACKOFF, backoff * 2)
            time.sleep(backoff)

def start_token_refresh_thread_if_needed() -> None:
    if st.session_state.get("token_refresher_started"):
        return
    # Require a refresh_token; fall back to header if proxy forwards it
    if not st.session_state.get("refresh_token"):
        # Optional header capture if proxy forwards refresh token
        headers = getattr(st.context, "headers", {}) or {}
        rt = headers.get("x-streamlit-refresh-token") or ""
        if rt:
            st.session_state.refresh_token = rt

    if not st.session_state.get("refresh_token"):
        # No refresh path; do not start thread
        st.session_state.token_refresher_started = True
        return

    st.session_state.stop_token_refresher = False
    th = threading.Thread(target=_token_refresher, name="token_refresher", daemon=True)
    add_script_run_ctx(th, get_script_run_ctx())
    th.start()
    st.session_state.token_refresher_started = True
    st.session_state.token_refresher_thread = th

def normalize(d: Dict[str, str]) -> Dict[str, str]:
    """Normalize dictionary keys and values to lowercase."""
    return {(k or "").strip().lower(): (v or "").strip() for k, v in (d or {}).items()}


def get_bearer_token(headers: Dict[str, str]) -> Optional[str]:
    """Extract bearer token from various header formats."""
    for name in ("authorization", "x-forwarded-access-token", "x-auth-request-access-token"):
        val = headers.get(name)
        if not val:
            continue
        return strip_bearer(val)
    return None


def strip_bearer(value: str) -> str:
    """Remove 'Bearer ' prefix from token."""
    v = (value or "").strip()
    if v.lower().startswith("bearer "):
        return v.split(" ", 1)[1].strip()
    return v


def get_user_info(access_token):
    """Get user info from Keycloak using access token."""
    keycloak_url = os.getenv("KEYCLOAK_URL")
    realm_name = os.getenv("KEYCLOAK_REALM")

    if not keycloak_url or not realm_name:
        return None

    userinfo_url = f"{keycloak_url}/realms/{realm_name}/protocol/openid-connect/userinfo"

    try:
        headers = {"Authorization": f"Bearer {access_token}"}
        response = requests.get(userinfo_url, headers=headers, timeout=10)
        response.raise_for_status()
        return response.json()
    except Exception as e:
        logger.error(f"Failed to get user info: {e}")
        return None


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


# -------------------------
# Session state initialization
# -------------------------
def init_session_state() -> None:
    # Ensure all keys exist; never assume presence across reruns
    if "authenticated" not in st.session_state:
        st.session_state.authenticated = False
    if "auth_method" not in st.session_state:
        st.session_state.auth_method = ""
    if "user_id" not in st.session_state:
        st.session_state.user_id = ""
    if "user_email" not in st.session_state:
        st.session_state.user_email = ""
    if "user_username" not in st.session_state:
        st.session_state.user_username = ""
    if "permissions" not in st.session_state:
        st.session_state.permissions = {}
    if "user_info" not in st.session_state:
        st.session_state.user_info = {"sub": "", "email": "", "preferred_username": ""}


# -------------------------
# Header utilities
# -------------------------
def normalize_headers(h: Dict[str, str]) -> Dict[str, str]:
    # Case-insensitive access
    return {(k or "").strip().lower(): (v or "").strip() for k, v in (h or {}).items()}


def bearer_from_headers(h: Dict[str, str]) -> Optional[str]:
    # Prefer Authorization, fallback to X-Forwarded-Access-Token, X-Auth-Request-Access-Token
    for name in ("authorization", "x-forwarded-access-token", "x-auth-request-access-token"):
        v = h.get(name)
        if not v:
            continue
        v = v.strip()
        if v.lower().startswith("bearer "):
            return v.split(" ", 1)[1].strip()
        return v
    return None


def decode_jwt_claims_no_verify(token: str) -> Dict:
    # Proxy already validated upstream; here we only need claims to hydrate identity
    try:
        return jwt.decode(token, options={"verify_signature": False, "verify_exp": False})
    except Exception as e:
        logger.warning(f"JWT decode (no verify) failed: {e}")
        return {}


# -------------------------
# Authentication
# -------------------------
def authenticate_from_proxy_or_jwt() -> None:
    # If already authenticated in this session, do not re-evaluate
    if st.session_state.authenticated:
        return

    headers = getattr(st.context, "headers", {}) or {}
    h = normalize_headers(headers)
    print("Headers h", h)
    print("Headers", headers)

    # Try Streamlit identity headers from proxy
    user_id = (
        h.get("x-streamlit-user-id")
        or headers.get("X-Streamlit-User-ID", "")
        or headers.get("X-Streamlit-User-Id", "")
        or ""
    )
    user_email = (
        h.get("x-streamlit-user-email")
        or headers.get("X-Streamlit-User-Email", "")
        or ""
    )
    user_username = (
        h.get("x-streamlit-user-username")
        or headers.get("X-Streamlit-User-Username", "")
        or ""
    )
    auth_method = (
        h.get("x-streamlit-auth-method")
        or headers.get("X-Streamlit-Auth-Method", "")
        or ""
    )
    perms_raw = (
        h.get("x-streamlit-user-permissions")
        or headers.get("X-Streamlit-User-Permissions", "")
        or ""
    )

    # If user_id empty, attempt to derive from JWT claims present in headers
    # Works for iframe flow (JWT) and for session flow when proxy added Authorization bearer
    if not user_id:
        token = bearer_from_headers(h)
        print("Token", token)
        if token:
            claims = decode_jwt_claims_no_verify(token)
            user_id = claims.get("sub") or user_id
            user_email = claims.get("email") or user_email
            user_username = claims.get("preferred_username") or user_username
            if not auth_method:
                auth_method = "jwt"

    # Fallback: if still no user_id but email exists, use email as stable identifier
    if not user_id and user_email:
        user_id = user_email

    # Parse permissions JSON safely
    permissions = {}
    if perms_raw:
        try:
            permissions = json.loads(perms_raw)
        except Exception:
            permissions = {}

    # Hydrate session state
    if user_id:
        st.session_state.authenticated = True
        st.session_state.auth_method = auth_method or ("session" if not (bearer_from_headers(h)) else "jwt")
        st.session_state.user_id = user_id
        st.session_state.user_email = user_email
        st.session_state.user_username = user_username or (user_email.split("@")[0] if user_email else "")
        st.session_state.permissions = permissions
        st.session_state.user_info = {
            "sub": st.session_state.user_id,
            "email": st.session_state.user_email,
            "preferred_username": st.session_state.user_username,
        }
        token_from_header = bearer_from_headers(h)  # already available

        st.session_state.access_token = token_from_header
        st.session_state.token_exp = _decode_exp_no_verify(token_from_header)

        rt_hdr = h.get("x-streamlit-refresh-token")
        if rt_hdr:
            st.session_state.refresh_token = rt_hdr

        # Kick off background refresh if a refresh token is present
        start_token_refresh_thread_if_needed()

        logger.info(
            f"Authenticated via {st.session_state.auth_method} as "
            f"{st.session_state.user_email or st.session_state.user_id}"
        )
    else:
        # Not authenticated: leave session_state.authenticated as False
        pass


# -------------------------
# App bootstrap
# -------------------------
init_session_state()
authenticate_from_proxy_or_jwt()

# Fail closed if not authenticated; Streamlit reruns on interactions so session_state persists per session
if not st.session_state.authenticated:
    st.error("❌ Authentication Error: Missing user information.")
    st.info("Please access this application through the main portal.")
    st.stop()

if __name__ == '__main__':
    from lex.lex_app.settings import repo_name

    try:
        exec(f"import {repo_name}._streamlit_structure as streamlit_structure")

        # Your existing model rendering logic...
        params = st.query_params
        model = params.get("model")
        pk = params.get("pk")

        if model and pk:
            # Instance-level visualization
            try:
                from django.apps import apps
                from lex.lex_app.settings import repo_name

                model_class = apps.get_model(repo_name, model)
                model_obj = model_class.objects.filter(pk=pk).first()

                if model_obj is None:
                    st.error(f"❌ Object with ID {pk} not found")
                elif not hasattr(model_obj, 'streamlit_main'):
                    st.error(f"❌ This model doesn't support visualization")
                else:
                    # Pass user info from session state
                    user = st.session_state.get('user_info')
                    model_obj.streamlit_main(user)

            except LookupError:
                st.error(f"❌ Model '{model}' not found")
            except Exception as e:
                st.error(f"❌ Error: {str(e)}")

        elif model and not pk:
            # Class-level visualization
            try:
                from django.apps import apps
                from lex.lex_app.settings import repo_name

                model_class = apps.get_model(repo_name, model)

                if not hasattr(model_class, 'streamlit_class_main'):
                    st.error(f"❌ This model doesn't support class-level visualization")
                else:
                    # Pass user info and permissions from session state
                    user = st.session_state.get('user_info')
                    permissions = st.session_state.get('permissions')
                    model_class.streamlit_class_main()

            except LookupError:
                st.error(f"❌ Model '{model}' not found")
            except Exception as e:
                st.error(f"❌ Error: {str(e)}")

        else:
            # Default application structure
            streamlit_structure.main()
            if st.button("Logout"):
                auth_method = st.session_state.get('auth_method', 'session')

                if auth_method == 'jwt':
                    # For JWT auth, just clear session and show message
                    st.session_state.stop_token_refresher = True
                    th = st.session_state.get("token_refresher_thread")
                    if th and th.is_alive():
                        # Give it a moment to exit; thread is daemon, so app exit also cleans up
                        th.join(timeout=1.0)
                    st.session_state.clear()
                    st.success("✅ Logged out successfully. You can close this window.")
                    st.stop()
                else:
                    # For session auth, redirect to logout
                    rd = urllib.parse.quote("http://localhost:8501", safe="")
                    st.markdown(
                        f"<meta http-equiv='refresh' content='0;url=/oauth2/sign_out?rd={rd}'>",
                        unsafe_allow_html=True
                    )

        # Logout functionality (adjust based on auth method)


    except Exception as e:
        if os.getenv("DEPLOYMENT_ENVIRONMENT") != "PROD":
            raise e
        else:
            with st.expander(":red[An error occurred while trying to load the app.]"):
                st.error(traceback.format_exc())


from django.http import JsonResponse
from django.views import View

from lex.audit_logging.handlers.lex_logger import LexLogger


class LexLoggerView(View):
    """
    A view to handle fetching logs. This assumes logs are stored in a 
    structured format like a database or a file you can read from.
    """

    def get(self, request, *args, **kwargs):
        # Here, you would fetch the logs from your storage (e.g., a database)
        # For simplicity, we are returning a static list of logs
        logs = [
            {'level': 'INFO', 'message': 'System started', 'timestamp': '2024-09-17 12:00:00'},
            {'level': 'WARNING', 'message': 'CPU usage high', 'timestamp': '2024-09-17 12:05:00'},
            {'level': 'ERROR', 'message': 'Application crashed', 'timestamp': '2024-09-17 12:10:00'},
        ]

        return JsonResponse(logs, safe=False)

    def post(self, request, *args, **kwargs):
        # Handle POST request to create a log (if needed)
        # Example of log creation or storage (typically, logging to file or database)
        message = request.POST.get('message', 'No message provided')
        log_level = request.POST.get('level', 'INFO')

        # Using the Python logging module to log
        logger = LexLogger()
        getattr(logger, log_level.lower(), logger.info)(message)

        return JsonResponse({'status': 'Log added successfully'}, status=201)



# core/management/commands/scan_model_changes.py
from django.core.management.base import BaseCommand
from django.apps import apps
from django.db.migrations.questioner import MigrationQuestioner
from django.db.migrations.state import ProjectState
from django.db.migrations.loader import MigrationLoader
from django.db.migrations.autodetector import MigrationAutodetector
from django.db.migrations.operations.models import CreateModel, DeleteModel, RenameModel

class Command(BaseCommand):
    help = "Scan pending model-level changes (add/delete/rename) before writing migrations."

    def handle(self, *args, **opts):
        # Mirror makemigrations flow but stop at the diff
        questioner = MigrationQuestioner(defaults={
            "ask_rename_model": True,  # auto-confirm model renames
        })

        loader = MigrationLoader(None, ignore_no_migrations=True)
        autodetector = MigrationAutodetector(
            loader.project_state(),
            ProjectState.from_apps(apps),
            questioner=questioner,

        )
        changes = autodetector.changes(graph=loader.graph)

        adds, deletes, renames = [], [], []
        for app_label, migrations in changes.items():
            for mig in migrations:
                for op in mig.operations:
                    if isinstance(op, CreateModel):
                        adds.append((app_label, op.name))
                    elif isinstance(op, DeleteModel):
                        deletes.append((app_label, op.name))
                    elif isinstance(op, RenameModel):
                        renames.append((app_label, op.old_name, op.new_name))

        # Call custom logic with only names, no file parsing
        # e.g., print or invoke callbacks
        for app, name in adds:
            self.stdout.write(f"ADD {app}.{name}")
        for app, name in deletes:
            self.stdout.write(f"DELETE {app}.{name}")
        for app, old, new in renames:
            self.stdout.write(f"RENAME {app}.{old} -> {new}")



import logging
from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer

class WebSocketHandler(logging.Handler):
    """
    A logging.Handler which, on each record, sends the formatted
    message to the appropriate WebSocket group (calculation_record).
    """
    def emit(self, record: logging.LogRecord):
        try:
            # format the message (use whatever Formatter you like)
            formatted = self.format(record)

            # pull our extra fields out of the record
            calc_record = record.__dict__.get("calculation_record")
            calc_id     = record.__dict__.get("calculationId")

            if not calc_record:
                # nothing to do if no group name
                return

            payload = {
                "type": "calculation_log_real_time",
                "payload": formatted,
            }

            channel_layer = get_channel_layer()
            async_to_sync(channel_layer.group_send)(
                f"{calc_record}",
                payload
            )
        except Exception:
            self.handleError(record)


from lex.core.calculated_updates.objects_to_recalculate_store import ObjectsToRecalculateStore
from lex.core.calculated_updates.update_handler import CalculatedModelUpdateHandler

__all__ = ['ObjectsToRecalculateStore', 'CalculatedModelUpdateHandler']



from django.core.management.base import BaseCommand
from django.apps import apps

from keycloak import KeycloakOpenIDConnection, KeycloakUMA, KeycloakAdmin


class Command(BaseCommand):
    help = (
        "Register each Django model as a Keycloak UMA resource, "
        "create a client role per resource‐scope, "
        "and wire up role‐policies & scope‐permissions"
    )

    def handle(self, *args, **options):
        # 1) connect via client‐credentials
        conn = KeycloakOpenIDConnection(
            server_url="https://exc-testing.com",
            realm_name="lex",
            client_id="LEX_LOCAL_ENV",
            client_secret_key="O1dT6TEXjsQWbRlzVxjwfUnNHPnwDmMF",
            verify=False,
        )
        kc_uma = KeycloakUMA(connection=conn)
        kc_admin = KeycloakAdmin(connection=conn)

        # 2) your client's internal UUID (hard‐coded or lookup)
        client_uuid = "3e5eeafe-a3b3-469e-9db3-54cff7108d70"

        # 3) load existing UMA resource‐sets
        existing = kc_uma.resource_set_list()
        existing_by_name = {r["name"]: r for r in existing}

        # 4) define the six scopes
        scopes = ["list", "show", "create", "edit", "delete", "export"]

        for model in apps.get_models():
            res_name = f"{model._meta.app_label}.{model.__name__}"

            # — create or skip UMA resource‐set
            if res_name in existing_by_name:
                resource = existing_by_name[res_name]
                resource_id = resource.get("_id") or resource.get("id")
                self.stdout.write(f"✔ UMA resource exists: {res_name}")
            else:
                payload = {
                    "name": res_name,
                    "type": "urn:django:model",
                    "scopes": [{"name": s} for s in scopes],
                    "ownerManagedAccess": False,
                }
                created = kc_uma.resource_set_create(payload)
                resource_id = created.get("_id") or created.get("id")
                self.stdout.write(
                    self.style.SUCCESS(f"✨ Created UMA resource: {res_name}")
                )

            # — for each scope, create a client role + policy + scope‐permission
            for scope in scopes:
                # a) **Client role** name
                role_name = f"{res_name}:{scope}"

                # c) **Scope‐based permission** tying
                #    – this UMA resource,
                #    – this one scope,
                #    – to the policy above.
                perm_name = role_name
                permission_payload = {
                    "name": perm_name,
                    "type": "scope",
                    "logic": "POSITIVE",
                    "decisionStrategy": "UNANIMOUS",
                    "resources": [resource_id],
                    "scopes": [scope],
                    "policies": [],
                }
                kc_admin.create_client_authz_scope_permission(
                    client_id=client_uuid, payload=permission_payload
                )
                self.stdout.write(
                    self.style.SUCCESS(f"🛡 Created scope permission: {perm_name}")
                )


from lex.utilities.config.generic_app_config import GenericAppConfig
from lex_app.apps import LexAppConfig


class UtilitiesConfig(LexAppConfig):
    """
    Configuration for the utilities app containing shared utilities, decorators, and helper functions.
    """
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'lex.utilities'
    verbose_name = 'Utilities'


import re
from functools import lru_cache
from rest_framework import serializers
from rest_framework.exceptions import PermissionDenied


def _camel_to_snake(name: str) -> str:
    """
    Converts a camelCase string to snake_case.
    For example: 'reportDate' -> 'report_date'
    """
    name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()


class PermissionAwareSerializerMixin:
    """
    A serializer mixin that injects field-level write-permission checks.

    This mixin intelligently handles permission validation with several key features:
    - Change Detection: Only checks permissions for fields where the incoming
      value is different from the value currently stored in the model instance.
    - Convention-Based Field Name Translation: Converts incoming camelCase field
      names to the snake_case names expected by the backend model.
    - Respects Model's Authority: Calls the `can_edit()` method on the model
      instance to get the authoritative set of editable fields.
    - Ignores Non-Editable & Special Fields: Automatically ignores fields that
      should not be part of a permission check, such as system-managed fields
      (`created_by`), reserved fields, and serializer method fields.
    - Pre-Validation Check: Hooks into `run_validation` to ensure permissions
      are checked before any other validation logic runs.
    """

    @lru_cache
    def _get_model_field_names(self) -> set:
        """
        Returns a set of all actual field names defined on the model.
        """
        if not hasattr(self.Meta, 'model'):
            return set()
        model = self.Meta.model
        return {f.name for f in model._meta.get_fields()}


    @lru_cache
    def _get_non_editable_fields(self) -> set:
        """
        Inspects the serializer's model and returns a set of field names that
        are not user-editable (e.g., marked with `editable=False`).
        """
        if not hasattr(self.Meta, 'model'):
            return set()
        model = self.Meta.model
        non_editable = {
            f.name for f in model._meta.get_fields()
            if hasattr(f, 'editable') and not f.editable
        }
        if model._meta.pk:
            non_editable.add(model._meta.pk.name)
        return non_editable

    def run_validation(self, data):
        """
        This is the main entry point for the permission check, overriding
        DRF's default to perform checks *before* field validation.
        """
        from rest_framework.exceptions import PermissionDenied
        from rest_framework import serializers
        
        request = self.context.get('request')
        instance = self.instance

        if instance and request and data:
            # Get editable fields using new system if available
            editable_backend_fields = set()
            
            try:
                if hasattr(instance, 'permission_edit'):
                    from lex.core.models.base import UserContext
                    user_context = UserContext.from_request(request, instance)
                    result = instance.permission_edit(user_context)
                    if result.allowed:
                        all_fields = {f.name for f in instance._meta.fields}
                        editable_backend_fields = result.get_fields(all_fields)
                elif hasattr(instance, 'can_edit'):
                    editable_backend_fields = instance.can_edit(request)
                    if not isinstance(editable_backend_fields, set):
                        editable_backend_fields = set()
                else:
                    # Default to allow all fields if no permission method
                    editable_backend_fields = {f.name for f in instance._meta.fields}
            except Exception:
                # Allow all fields by default on error
                editable_backend_fields = {f.name for f in instance._meta.fields}

            if not editable_backend_fields:
                raise PermissionDenied(
                    f"You do not have permission to edit this {instance.__class__.__name__}."
                )

            model_field_names = self._get_model_field_names()
            non_editable_fields = self._get_non_editable_fields()

            for frontend_field_name, new_raw_value in data.items():
                if frontend_field_name.startswith('lexReserved'):
                    continue

                backend_field_name = _camel_to_snake(frontend_field_name)

                if backend_field_name not in model_field_names:
                    continue
                if backend_field_name in non_editable_fields:
                    continue

                # --- CHANGE DETECTION LOGIC ---
                field_obj = self.fields.get(frontend_field_name)
                if field_obj:
                    try:
                        # Convert incoming value to a Python object for accurate comparison
                        new_python_value = field_obj.to_internal_value(new_raw_value)
                        old_python_value = getattr(instance, backend_field_name)

                        # If values are the same, no change was made. Skip permission check.
                        if new_python_value == old_python_value:
                            continue
                    except (serializers.ValidationError, AttributeError):
                        # If conversion fails or attribute doesn't exist, validation
                        # will fail later. For now, we must assume it's an attempted
                        # change and check permission.
                        pass
                # --- END OF CHANGE DETECTION ---

                # If we reach here, it's a changed field. Now, check permission.
                if backend_field_name not in editable_backend_fields:
                    raise PermissionDenied(
                        f"You do not have permission to edit the '{frontend_field_name}' field."
                    )

        elif not instance and request:
            model_class = self.Meta.model
            can_create = False
            
            try:
                temp_instance = model_class()
                if hasattr(temp_instance, 'permission_create'):
                    from lex.core.models.base import UserContext
                    user_context = UserContext.from_request(request, temp_instance)
                    can_create = temp_instance.permission_create(user_context)
                elif hasattr(temp_instance, 'can_create'):
                    can_create = temp_instance.can_create(request)
                else:
                    # Default to allow if no permission method
                    can_create = True
            except Exception:
                # Allow by default on error
                can_create = True
                
            if not can_create:
                raise PermissionDenied(
                    f"You do not have permission to create a {model_class.__name__}."
                )

        return super().run_validation(data)


def add_permission_checks(serializer_class):
    """
    A class decorator that injects permission-checking capabilities into any
    ModelSerializer.
    """
    class PermissionAwareVersion(PermissionAwareSerializerMixin, serializer_class):
        pass

    PermissionAwareVersion.__name__ = serializer_class.__name__
    PermissionAwareVersion.__module__ = serializer_class.__module__
    return PermissionAwareVersion


# Metaclass approach for automatic permission injection
class PermissionAwareSerializerMetaclass(type(serializers.ModelSerializer)):
    """
    Metaclass that automatically adds permission checking to all model serializers.
    """

    def __new__(mcs, name, bases, attrs):
        # Don't modify the base classes themselves
        if name in ('PermissionAwareSerializerMixin', 'LexModelSerializer'):
            return super().__new__(mcs, name, bases, attrs)

        # Check if this is a ModelSerializer for a LexModel
        meta_class = attrs.get('Meta')
        if meta_class and hasattr(meta_class, 'model'):
            model = meta_class.model
            # Check if the model inherits from LexModel
            if hasattr(model, 'can_read') and hasattr(model, 'can_edit'):
                # Inject the permission mixin if not already present
                if PermissionAwareSerializerMixin not in bases:
                    bases = (PermissionAwareSerializerMixin,) + bases

        return super().__new__(mcs, name, bases, attrs)




from django.core.management.base import BaseCommand

from keycloak import KeycloakOpenIDConnection, KeycloakUMA, KeycloakAdmin
from keycloak.exceptions import KeycloakGetError


class Command(BaseCommand):
    help = (
        "Register each Django model as a Keycloak UMA resource, "
        "create a client role per resource-scope, "
        "and wire up role-policies & resource-permissions"
    )

    def handle(self, *args, **options):
        # 1) connect via client-credentials
        conn = KeycloakOpenIDConnection(
            server_url="https://exc-testing.com",
            realm_name="lex",
            client_id="LEX_LOCAL_ENV",
            client_secret_key="O1dT6TEXjsQWbRlzVxjwfUnNHPnwDmMF",
            verify=False,
        )
        kc_uma = KeycloakUMA(connection=conn)
        kc_admin = KeycloakAdmin(connection=conn)

        # 2) your client’s internal UUID (hard-coded or lookup)
        client_uuid = "3e5eeafe-a3b3-469e-9db3-54cff7108d70"
        role_name = "test"
        try:
            role = kc_admin.get_client_role(client_uuid, role_name)
            self.stdout.write(f"✔ Client role exists: {role_name}")
        except KeycloakGetError:
            # **correct** call signature: first arg is client_id
            kc_admin.create_client_role(client_uuid, {"name": role_name})
            role = kc_admin.get_client_role(client_uuid, role_name)
            self.stdout.write(self.style.SUCCESS(f"✨ Created client role: {role_name}"))



"""
Cache management utilities for the CalculationLog system.

This module provides the CacheManager class for handling cache operations
including message storage, cache cleanup, and graceful degradation when Cache
is unavailable.

Note on Cache Cleanup:
The cleanup_calculation() method attempts to find cache keys by pattern matching.
However, pattern-based key discovery can be unreliable depending on the
configuration and django-redis version. If pattern matching fails, the method
gracefully degrades by skipping cleanup - this is acceptable since cache entries
have a TTL and will expire naturally (default: 1 week).

For more reliable cleanup, use cleanup_specific_key() when you know the exact
cache keys, or pass specific_keys to cleanup_calculation().
"""

import logging
import os
from typing import List, Optional
from django.core.cache import caches
from django.core.cache.backends.base import InvalidCacheBackendError
from lex.audit_logging.utils.data_models import CacheCleanupResult, CacheOperationError

logger = logging.getLogger(__name__)


class CacheManager:
    """
    Manages cache operations for calculation logging.
    
    This class provides static methods for storing log messages in cache,
    cleaning up cache entries after calculations complete, and handling
    unavailability gracefully.
    """
    
    CACHE_TIMEOUT = 60 * 60 * 24 * 7  # Cache for one week
    CALC_CACHE_NAME = "redis" if os.getenv("DEPLOYMENT_ENVIRONMENT") else "default"

    @staticmethod
    def store_message(cache_key: str, message: str) -> bool:
        """
        Store log message in cache with error handling.
        
        Args:
            cache_key: The cache key to store the message under
            message: The log message to store
            
        Returns:
            bool: True if message was stored successfully, False otherwise
            
        Raises:
            CacheOperationError: If cache operation fails and graceful degradation is disabled
        """
        try:
            cache = caches[CacheManager.CALC_CACHE_NAME]
            # print("System cache:", caches)
            
            # Get existing message from cache, append new message
            existing_message = cache.get(cache_key, "")
            updated_message = existing_message + "\n" + message if existing_message else message
            
            # Store updated message with timeout
            cache.set(cache_key, updated_message)
            
            logger.debug(f"Successfully stored message in cache with key: {cache_key}")
            return True
            
        except InvalidCacheBackendError:
            logger.warning(f"Cache backend not available, skipping cache storage for key: {cache_key}")
            return False
            
        except Exception as e:
            logger.error(f"Failed to store message in cache with key {cache_key}: {str(e)}")
            # Graceful degradation - don't raise exception, just log and continue
            return False


    @staticmethod
    def cleanup_calculation(calculation_id: str = None, specific_keys: Optional[List[str]] = None) -> CacheCleanupResult:
        """
        Remove all cache entries for a completed calculation.
        
        Args:
            calculation_id: The calculation ID to clean up cache entries for
            specific_keys: Optional list of specific cache keys to clean up.
                          If provided, only these keys will be cleaned.
                          If None, will attempt to find keys by pattern matching.
            
        Returns:
            CacheCleanupResult: Results of the cleanup operation including
                               success status, cleaned keys, and any errors
        """
        cleaned_keys = []
        errors = []
        
        try:
            cache = caches[CacheManager.CALC_CACHE_NAME]

            
            # Determine which keys to clean up
            if specific_keys is not None:
                # Use provided specific keys
                pattern_keys = specific_keys
                logger.debug(f"Cleaning up {len(specific_keys)} specific cache keys for calculation {calculation_id}")
            elif calculation_id is not None:
                # Try to find keys by pattern matching
                pattern_keys = CacheManager._find_calculation_keys(cache, calculation_id)
                logger.debug(f"Found {len(pattern_keys)} cache keys by pattern matching for calculation {calculation_id}")
            else:
                return CacheCleanupResult(success=True, errors=errors, cleaned_keys=cleaned_keys)
            
            for key in pattern_keys:
                try:
                    cache.delete(key)
                    cleaned_keys.append(key)
                    logger.debug(f"Successfully cleaned cache key: {key}")
                except Exception as e:
                    error_msg = f"Failed to delete cache key {key}: {str(e)}"
                    errors.append(error_msg)
                    logger.error(error_msg)
            
            success = len(errors) == 0
            logger.info(f"Cache cleanup for calculation {calculation_id} completed. "
                       f"Cleaned {len(cleaned_keys)} keys, {len(errors)} errors")
            
            return CacheCleanupResult(
                success=success,
                cleaned_keys=cleaned_keys,
                errors=errors
            )
            
        except InvalidCacheBackendError:
            error_msg = "Cache backend not available for cleanup"
            logger.warning(error_msg)
            return CacheCleanupResult(
                success=False,
                cleaned_keys=[],
                errors=[error_msg]
            )
            
        except Exception as e:
            error_msg = f"Unexpected error during cache cleanup for calculation {calculation_id}: {str(e)}"
            logger.error(error_msg)
            return CacheCleanupResult(
                success=False,
                cleaned_keys=cleaned_keys,
                errors=errors + [error_msg]
            )
    
    @staticmethod
    def build_cache_key(calculation_record: str, calc_id: str) -> str:
        """
        Generate cache key using pattern {calculation_record}_{calc_id}.
        
        Args:
            calculation_record: The calculation record identifier (e.g., "model_name_pk")
            calc_id: The calculation ID
            
        Returns:
            str: The formatted cache key
        """
        if not calculation_record or not calc_id:
            raise ValueError("Both calculation_record and calc_id must be provided")
        
        return f"{calculation_record}_{calc_id}"
    
    @staticmethod
    def _find_calculation_keys(cache, calculation_id: str) -> List[str]:
        """
        Find all cache keys associated with a calculation ID.
        
        This method attempts to find keys using available django-redis methods.
        Since pattern-based key discovery can be unreliable, this method gracefully
        handles failures and returns an empty list when keys cannot be found.
        
        Args:
            cache: The cache instance
            calculation_id: The calculation ID to search for
            
        Returns:
            List[str]: List of cache keys associated with the calculation
        """
        try:
            # For django-redis, try to use the iter_keys method if available
            # This is more reliable than trying to access the underlying client
            if hasattr(cache, 'iter_keys'):
                pattern = f"*_{calculation_id}"
                matching_keys = list(cache.iter_keys(pattern))
                return [str(key) for key in matching_keys if key]
            
            # Fallback: Try the keys method
            elif hasattr(cache, 'keys'):
                pattern = f"*_{calculation_id}"
                matching_keys = cache.keys(pattern)
                return [str(key) for key in matching_keys if key]
            
            else:
                # If no pattern matching is available, log and return empty list
                logger.info(f"Pattern-based key cleanup not available for calculation {calculation_id}. "
                           "Cache entries will expire naturally based on TTL.")
                return []
            
        except Exception as e:
            logger.warning(f"Could not retrieve cache keys for pattern matching: {str(e)}")
            # Graceful degradation: return empty list
            # Cache entries will expire naturally based on their TTL (1 week)
            # This is acceptable behavior as cache cleanup is an optimization, not a requirement
            return []
    
    @staticmethod
    def get_message(cache_key: str) -> Optional[str]:
        """
        Retrieve a message from cache.
        
        Args:
            cache_key: The cache key to retrieve
            
        Returns:
            Optional[str]: The cached message if found, None otherwise
        """
        try:
            cache = caches[CacheManager.CALC_CACHE_NAME]
            return cache.get(cache_key)
            
        except InvalidCacheBackendError:
            logger.warning(f"Cache backend not available, cannot retrieve key: {cache_key}")
            return None
            
        except Exception as e:
            logger.error(f"Failed to retrieve message from cache with key {cache_key}: {str(e)}")
            return None
    
    @staticmethod
    def is_cache_available() -> bool:
        """
        Check if cache is available.
        
        Returns:
            bool: True if cache is available, False otherwise
        """
        try:
            cache = caches[CacheManager.CALC_CACHE_NAME]
            # Try a simple operation to test connectivity
            cache.set("__cache_test__", "test", timeout=1)
            cache.delete("__cache_test__")
            return True
            
        except Exception:
            return False
    
    @staticmethod
    def cleanup_specific_key(cache_key: str) -> bool:
        """
        Clean up a specific cache key.
        
        This method is useful when you know the exact cache key to clean up,
        avoiding the need for pattern matching.
        
        Args:
            cache_key: The specific cache key to delete
            
        Returns:
            bool: True if the key was successfully deleted, False otherwise
        """
        try:
            cache = caches[CacheManager.CALC_CACHE_NAME]
            cache.delete(cache_key)
            logger.debug(f"Successfully deleted cache key: {cache_key}")
            return True
            
        except InvalidCacheBackendError:
            logger.warning(f"Cache backend not available, cannot delete key: {cache_key}")
            return False
            
        except Exception as e:
            logger.error(f"Failed to delete cache key {cache_key}: {str(e)}")
            return False


# setup.py
import os
import shutil
import sys
import subprocess
from pathlib import Path

from setuptools import setup, find_packages
from setuptools.command.install import install

with open("requirements.txt") as f:
    install_requires = f.read().splitlines()

PROJECT_MARKERS = {'.git', 'pyproject.toml', 'setup.cfg', 'manage.py', 'requirements.txt', '.idea', '.vscode'}

def _resolve_project_root(default=None):
    base = Path(default or os.getcwd()).resolve()
    try:
        out = subprocess.run(
            ["git", "rev-parse", "--show-toplevel"],
            cwd=str(base),
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            text=True,
            check=True,
        )
        return out.stdout.strip()
    except Exception:
        pass
    for p in [base] + list(base.parents):
        if any((p / m).exists() for m in PROJECT_MARKERS):
            return str(p)
    return str(base)

class CustomInstallCommand(install):
    def run(self):
        install.run(self)
        # self.move_other_directory()
        # self.generate_pycharm_configs()
        # self.generate_env_file()

    def move_other_directory(self):
        source = os.path.join(os.path.dirname(__file__), 'lex', 'generic_app')
        target = os.path.join(os.path.dirname(self.install_lib), 'generic_app')
        if os.path.exists(target):
            shutil.rmtree(target)
        shutil.move(source, target)
        print(f'Moved other_directory to {target}')

    def _project_root(self):
        return _resolve_project_root(os.getcwd())

    def generate_env_file(self):
        project_root = self._project_root()
        env_path = os.path.join(project_root, '.env')
        if os.path.exists(env_path):
            print(f'.env already exists at {env_path}, not modifying.')
            return
        content = """KEYCLOAK_URL=https://auth.excellence-cloud.dev
KEYCLOAK_REALM=
OIDC_RP_CLIENT_ID=
OIDC_RP_CLIENT_SECRET=
OIDC_RP_CLIENT_UUID=
"""
        with open(env_path, 'w', encoding='utf-8') as f:
            f.write(content)
        print(f'Created .env at {env_path}')

    def generate_pycharm_configs(self):
        project_root = self._project_root()
        runconfigs_dir = os.path.join(project_root, '.run')
        os.makedirs(runconfigs_dir, exist_ok=True)
        project_name = os.path.basename(project_root)
        env_file_path = os.path.join(project_root, '.env')
        env_files_option = f'<option name="ENV_FILES" value="{env_file_path}" />' if os.path.exists(env_file_path) else '<option name="ENV_FILES" value="" />'
        configs = {
            'Init.run.xml': {'name': 'Init', 'parameters': 'Init'},
            'Start.run.xml': {'name': 'Start', 'parameters': 'start --reload --loop asyncio lex_app.asgi:application'},
            'Make_migrations.run.xml': {'name': 'Make migrations', 'parameters': 'makemigrations'},
            'Migrate.run.xml': {'name': 'Migrate', 'parameters': 'migrate'},
            'Streamlit.run.xml': {'name': 'Streamlit', 'parameters': 'streamlit run streamlit_app.py'},
            'Create_DB.run.xml': {'name': 'Create DB', 'parameters': 'test lex.lex_app.logging.create_db.create_db --keepdb'},
            'Flush_DB.run.xml': {'name': 'Flush DB', 'parameters': 'flush'},
        }
        for filename, config in configs.items():
            config_content = f'''<component name="ProjectRunConfigurationManager">
  <configuration default="false" name="{config['name']}" type="PythonConfigurationType" factoryName="Python">
    <module name="{project_name}" />
    {env_files_option}
    <option name="INTERPRETER_OPTIONS" value="" />
    <option name="PARENT_ENVS" value="true" />
    <envs>
      <env name="PYTHONUNBUFFERED" value="1" />
    </envs>
    <option name="SDK_HOME" value="" />
    <option name="WORKING_DIRECTORY" value="{project_root}" />
    <option name="IS_MODULE_SDK" value="true" />
    <option name="ADD_CONTENT_ROOTS" value="true" />
    <option name="ADD_SOURCE_ROOTS" value="true" />
    <EXTENSION ID="PythonCoverageRunConfigurationExtension" runner="coverage.py" />
    <option name="SCRIPT_NAME" value="lex" />
    <option name="PARAMETERS" value="{config['parameters']}" />
    <option name="SHOW_COMMAND_LINE" value="false" />
    <option name="EMULATE_TERMINAL" value="false" />
    <option name="MODULE_MODE" value="true" />
    <option name="REDIRECT_INPUT" value="false" />
    <option name="INPUT_FILE" value="" />
    <method v="2" />
  </configuration>
</component>'''
            config_path = os.path.join(runconfigs_dir, filename)
            with open(config_path, 'w', encoding='utf-8') as f:
                f.write(config_content)
            print(f'Generated PyCharm run configuration: {config_path}')
        print(f'PyCharm run configurations generated in: {runconfigs_dir}')
        if os.path.exists(env_file_path):
            print(f'Configurations will use .env file: {env_file_path}')
        else:
            print(f'No .env file found at {env_file_path}. Create one if needed for environment variables.')

setup(
    name="lex-app",
    version="2.0.0rc8",
    author="Melih Sünbül",
    author_email="m.sunbul@excellence-cloud.com",
    description="A Python / Django library to create business applications easily with complex logic",
    long_description_content_type="text/markdown",
    url="https://github.com/ExcellenceCloudGmbH/lex-app",
    packages=find_packages(),
    include_package_data=True,
    py_modules=['generate_pycharm_configs'],
    entry_points={
        "console_scripts": [
            "lex = lex.__main__:main",
            "lex-generate-configs = generate_pycharm_configs:generate_pycharm_configs",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=install_requires,
    python_requires=">=3.6",
    cmdclass={'install': CustomInstallCommand},
)



from __future__ import absolute_import

import sys
import warnings
from pathlib import Path

from django.core.cache import CacheKeyWarning
from google.oauth2 import service_account

"""
Django settings for lex_app project.

Generated by 'django-admin startproject' using Django 3.0.8.

For more information on this file, see
https://docs.djangoproject.com/en/3.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.0/ref/settings/
"""

import os

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
from datetime import timedelta

import sentry_sdk
from sentry_sdk.integrations.django import DjangoIntegration


def traces_sampler(sampling_context):
    if sampling_context == "/health":
        # Drop this transaction, by setting its sample rate to 0%
        return 0
    else:
        # Default sample rate for all others (replaces traces_sample_rate)
        return 0.1


if (
    os.getenv("DEPLOYMENT_ENVIRONMENT") != "DEV"
    and os.getenv("DEPLOYMENT_ENVIRONMENT") is not None
):
    sentry_sdk.init(
        dsn="https://a3aa24a7fccd42dbbb7457e6402b5443@o1318244.ingest.sentry.io/6573603",
        integrations=[
            DjangoIntegration(),
        ],
        environment=os.getenv("DEPLOYMENT_ENVIRONMENT"),
        # Set traces_sample_rate to 1.0 to capture 100%
        # of transactions for performance monitoring.
        # We recommend adjusting this value in production.
        traces_sampler=traces_sampler,
        # If you wish to associate users to errors (assuming you are using
        # django.contrib.auth) you may enable sending PII data.
        send_default_pii=True,
    )


warnings.simplefilter("ignore", CacheKeyWarning)

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
NEW_BASE_DIR = Path(os.getenv("PROJECT_ROOT", os.getcwd())).parent.as_posix()
sys.path.append(NEW_BASE_DIR)

ASGI_APPLICATION = "lex_app.asgi.application"

if os.getenv("DEPLOYMENT_ENVIRONMENT") is None:
    CHANNEL_LAYERS = {
        "default": {
            "BACKEND": "channels.layers.InMemoryChannelLayer",
        },
    }
else:
    CHANNEL_LAYERS = {
        "default": {
            "BACKEND": "channels_redis.pubsub.RedisPubSubChannelLayer",
            "CONFIG": {
                "hosts": [
                    (
                        f"redis://{os.getenv('REDIS_USERNAME')}:{os.getenv('REDIS_PASSWORD')}@{os.getenv('REDIS_HOST')}/2"
                        if os.getenv("DEPLOYMENT_ENVIRONMENT") is not None
                        else "redis://127.0.0.1:6379/2"
                    )
                ],
                "capacity": 100000,
                "expiry": 10,
                "prefix": f"{os.getenv('INSTANCE_RESOURCE_IDENTIFIER', 'local')}:",
            },
        },
    }

STORAGES = {
    "default": {
        "BACKEND": "lex.utilities.storage.custom_storage.CustomDefaultStorage",
    },
    "staticfiles": {"BACKEND": "django.contrib.staticfiles.storage.StaticFilesStorage"},

}

SHAREPOINT_APP_CLIENT_ID = os.getenv("SHAREPOINT_APP_CLIENT_ID")
SHAREPOINT_API_CERTIFICATE_PATH = os.getenv("SHAREPOINT_API_CERTIFICATE_PATH")
SHAREPOINT_API_CERTIFICATE_THUMBPRINT = os.getenv(
    "SHAREPOINT_API_CERTIFICATE_THUMBPRINT"
)
SHAREPOINT_API_TENANT_NAME = os.getenv("SHAREPOINT_API_TENANT_NAME")
SHAREPOINT_URL = os.getenv("SHAREPOINT_URL", "local")
SHAREPOINT_STATIC_DIR = "static"
SHAREPOINT_MEDIA_DIR = "uploads"

if os.getenv("STORAGE_TYPE") == "SHAREPOINT":
    STORAGES = {
        "default": {
            "BACKEND": "django_sharepoint_storage.SharePointCloudStorageUtils.Media",
        }
    }
    MEDIA_ROOT = "uploads/"

if os.getenv("STORAGE_TYPE") == "GCS":
    STORAGES = {
        "default": {
            "BACKEND": "lex.utilities.utils.gcs_utils.Media",
        }
    }
    GS_BUCKET_NAME = os.getenv("GS_BUCKET_NAME")
    GS_CREDENTIALS = service_account.Credentials.from_service_account_file(
        os.path.join(NEW_BASE_DIR, "django-storages", "gcpCredentials.json"),
    )
    MEDIA_ROOT = "uploads/"

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.getenv(
    "DJANGO_SECRET_KEY", default="pjlulvaa77lteno-_y6!oxb%63xqiaw4%n%1or&77a!x9@nkd+"
)

# SECURITY WARNING: don't run with debug turned on in production!
if os.getenv("LEX_ENVIRONMENT_TAG", "envvar_not_existing") == "dev":
    DEBUG = True
else:
    DEBUG = False

ALLOWED_HOSTS = [
    os.getenv("DOMAIN_HOSTED", "localhost"),
    "127.0.0.1",
    "localhost",
    os.environ.get("POD_IP", default="envvar_not_existing"),
]

LOGIN_REDIRECT_URL = "/process_admin/all"

CSRF_TRUSTED_ORIGINS = [
    "https://*." + os.getenv("DOMAIN_HOSTED", "localhost"),
    "http://localhost:3000",
]

REACT_APP_BUILD_PATH = (
    Path(__file__).resolve().parent.parent / Path("react/build")
).as_posix()
repo_name = os.getenv("PROJECT_ROOT", Path(os.getcwd()).resolve().as_posix()).split(
    "/"
)[-1]
LEGACY_MEDIA_ROOT = os.path.join(NEW_BASE_DIR, f"{repo_name}/")
LOG_FILE_PATH = os.path.join(NEW_BASE_DIR, f"{repo_name}/{repo_name}.log")


LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "verbose": {
            "format": "{levelname} {asctime} {module} {process:d} {thread:d} {message}",
            "style": "{",
        },
    },
    "handlers": {
        "file": {
            "level": "DEBUG",
            "class": "logging.FileHandler",
            "filename": LOG_FILE_PATH,
            "formatter": "verbose",
        },
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "simple",
        },
        "websocket": {
            "level": "DEBUG",
            "class": "lex.audit_logging.handlers.websocket_handler.WebSocketHandler",
        },
    },
    "loggers": {
        "LexLogger": {
            "handlers": ["file"],
            "level": "DEBUG",
            "propagate": True,
        },
    },
}


# Application definition

INSTALLED_APPS = [
    "channels",
    "lex.utilities.apps.UtilitiesConfig",
    "lex.core.apps.CoreConfig",
    "lex.authentication.apps.AuthenticationConfig",
    "lex.audit_logging.apps.LoggingConfig",
    "lex.api.apps.ApiConfig",
    "lex.process_admin.apps.ProcessAdminConfig",
    "lex.lex_app.apps.LexAppConfig",
    "simple_history",
    "celery",
    "react",
    "markdown",
    "lex.lex_app.apps.CustomAdminConfig",  # Custom admin config to prevent auth admin conflicts
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "rest_framework",
    "rest_framework_api_key",
    "django.contrib.postgres",
    "django_db_views",
    "oauth2_authcodeflow",
]

if not repo_name.startswith("lex"):
    INSTALLED_APPS.append(repo_name)


CRISPY_FAIL_SILENTLY = not DEBUG

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "corsheaders.middleware.CorsMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "simple_history.middleware.HistoryRequestMiddleware",
    "lex.api.middleware.keycloak_permissions.KeycloakPermissionsMiddleware",
    "oauth2_authcodeflow.middleware.LoginRequiredMiddleware",
    "oauth2_authcodeflow.middleware.RefreshSessionMiddleware",
    "oauth2_authcodeflow.middleware.RefreshAccessTokenMiddleware",
]

ROOT_URLCONF = "lex_app.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [os.path.join(BASE_DIR, "templates"), "generic_app/submodels/"],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "lex_app.wsgi.application"

DISABLE_SERVER_SIDE_CURSORS = True

db_username = os.getenv("POSTGRES_USERNAME", "django")
kubernetes = not os.getenv("KUBERNETES_ENGINE", "NONE") == "NONE"

CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.db.DatabaseCache",
        "LOCATION": "default_cache",
        "TIMEOUT": None,
    },
    "oidc": {
        "BACKEND": "django.core.cache.backends.db.DatabaseCache",
        "LOCATION": "oidc_cache",
    },
    "redis": {
        "BACKEND": "django_redis.cache.RedisCache",
        "LOCATION": (
            f"redis://{os.getenv('REDIS_USERNAME')}:{os.getenv('REDIS_PASSWORD')}@{os.getenv('REDIS_HOST')}/2"
            if os.getenv("DEPLOYMENT_ENVIRONMENT") is not None
            else "redis://127.0.0.1:6379/2"
        ),
        "OPTIONS": {
            "CLIENT_CLASS": "django_redis.client.DefaultClient",
            # Optional: silently ignore redis down-time
            "IGNORE_EXCEPTIONS": True,
        },
    },
    "local": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
        "LOCATION": "unique-snowflake",
    }

}


DATABASES = {
    "local": {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, f"{repo_name}.sqlite3"),
    },
    "default": {
        "ENGINE": "django.db.backends.postgresql_psycopg2",
        "NAME": f"db_{repo_name}",
        "USER": "django",
        "PASSWORD": "lundadminlocal",
        "HOST": "localhost",
        "PORT": "5432",
    },
    "GCP": {
        "ENGINE": "django.db.backends.postgresql_psycopg2",
        "NAME": os.getenv("DATABASE_NAME", "envvar_not_existing"),
        "USER": db_username,
        "PASSWORD": os.getenv("POSTGRES_PASSWORD", "envvar_not_existing"),
        "HOST": os.getenv("DATABASE_DOMAIN", "envvar_not_existing"),
        "PORT": "5432",
        "TEST": {
            "NAME": os.getenv("DATABASE_NAME", "envvar_not_existing"),
        },
    },
    "DOCKER-COMPOSE": {
        "ENGINE": "django.db.backends.postgresql_psycopg2",
        "NAME": os.getenv("DATABASE_NAME", "envvar_not_existing"),
        "USER": db_username,
        "PASSWORD": os.getenv("POSTGRES_PASSWORD", "envvar_not_existing"),
        "HOST": os.getenv("DATABASE_DOMAIN", "envvar_not_existing"),
        "PORT": "5432",
        "TEST": {
            "NAME": os.getenv("DATABASE_NAME", "envvar_not_existing"),
        },
    },
    "K8S": {
        "ENGINE": "django.db.backends.postgresql_psycopg2",
        "NAME": os.getenv("DATABASE_NAME", "envvar_not_existing"),
        "USER": db_username,
        "PASSWORD": os.getenv("POSTGRES_PASSWORD", "envvar_not_existing"),
        "HOST": os.getenv("DATABASE_DOMAIN", "envvar_not_existing"),
        "PORT": "5432",
        "OPTIONS": {"sslmode": "disable"},
        "TEST": {
            "NAME": os.getenv("DATABASE_NAME", "envvar_not_existing"),
        },
    },
}
DATABASE_DEPLOYMENT_TARGET = os.getenv("DATABASE_DEPLOYMENT_TARGET", "local")
if DATABASE_DEPLOYMENT_TARGET != "local":
    DATABASES["default"] = DATABASES[DATABASE_DEPLOYMENT_TARGET]
else:
    DATABASES["default"] = DATABASES["local"]

MIGRATION_MODULES = {}


# CELERY STUFF
# Celery Broker Configuration - Redis
CELERY_BROKER_URL = (
    f"redis://{os.getenv('REDIS_USERNAME')}:{os.getenv('REDIS_PASSWORD')}@{os.getenv('REDIS_HOST')}/1"
    if os.getenv("DEPLOYMENT_ENVIRONMENT") is not None
    else "redis://127.0.0.1:6379/1"
)

# Celery Result Backend Configuration - Redis (changed from PostgreSQL)
CELERY_RESULT_BACKEND = (
    f"db+postgresql://{db_username}:{os.getenv('POSTGRES_PASSWORD', 'envvar_not_existing')}@{os.getenv('DATABASE_DOMAIN', 'envvar_not_existing')}/{os.getenv('DATABASE_NAME', 'envvar_not_existing')}"
    if os.getenv("DEPLOYMENT_ENVIRONMENT") is not None
    else f"db+postgresql://django:lundadminlocal@localhost/db_{repo_name}"
)

# Celery Configuration
CELERY_CACHE_BACKEND = "default"
CELERY_ACCEPT_CONTENT = ["application/json", "pickle"]
CELERY_TASK_SERIALIZER = "pickle"
CELERY_RESULT_SERIALIZER = "pickle"
CELERY_CREATE_MISSING_QUEUES = True
CELERY_TASK_TRACK_STARTED = True
CELERY_RESULT_PERSISTENT = True
CELERY_TASK_DEFAULT_QUEUE = os.getenv("INSTANCE_RESOURCE_IDENTIFIER", "celery")

# Redis-specific Celery configuration
CELERY_BROKER_TRANSPORT_OPTIONS = {
    "global_keyprefix": f"{os.getenv('INSTANCE_RESOURCE_IDENTIFIER', 'celery')}:",
    "visibility_timeout": float("inf"),
}
CELERY_RESULT_BACKEND_TRANSPORT_OPTIONS = {
    "global_keyprefix": f"{os.getenv('INSTANCE_RESOURCE_IDENTIFIER', 'celery')}:",
    "visibility_timeout": float("inf"),
}

# Task execution configuration
CELERY_TASK_ACKS_LATE = True
CELERY_TASK_REJECT_ON_WORKER_LOST = True
CELERY_WORKER_PREFETCH_MULTIPLIER = 1

# Result expiration (Redis cleanup)
CELERY_RESULT_EXPIRES = 3600  # 1 hour

# Enhanced Celery Active Configuration
# Support both CELERY_ACTIVE and legacy C_FORCE_ROOT for backward compatibility
CELERY_ACTIVE = os.getenv("CELERY_ACTIVE", "False").lower() == "true"
CELERY_TASK_TIMEOUT = int(os.getenv("CELERY_TASK_TIMEOUT", "3600"))  # 1 hour
CELERY_MAX_RETRIES = int(os.getenv("CELERY_MAX_RETRIES", "3"))
CELERY_RETRY_DELAY = int(os.getenv("CELERY_RETRY_DELAY", "60"))  # seconds

# Legacy support for C_FORCE_ROOT (backward compatibility)
try:
    c_force_root = os.getenv("C_FORCE_ROOT")
    if c_force_root == "True":
        CELERY_ACTIVE = True
    # If CELERY_ACTIVE is already True from environment, keep it True
    # Only set to False if both C_FORCE_ROOT and CELERY_ACTIVE are not set to True
    elif not CELERY_ACTIVE:
        CELERY_ACTIVE = False
except Exception as e:
    # If there's any error, fall back to the CELERY_ACTIVE environment variable value
    pass

# Make celery_active available as a module-level variable for backward compatibility
celery_active = CELERY_ACTIVE

# Password validation
# https://docs.djangoproject.com/en/3.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]

CORS_ORIGIN_ALLOW_ALL = True

DATA_UPLOAD_MAX_NUMBER_FIELDS = None
KEYCLOAK_URL = os.getenv("KEYCLOAK_URL")
KEYCLOAK_REALM = os.getenv("KEYCLOAK_REALM")
OIDC_RP_CLIENT_ID = os.getenv("OIDC_RP_CLIENT_ID")
OIDC_RP_CLIENT_SECRET = os.getenv("OIDC_RP_CLIENT_SECRET")
OIDC_RP_CLIENT_UUID = os.getenv("OIDC_RP_CLIENT_UUID")
OIDC_OP_USER_ENDPOINT          = f"{KEYCLOAK_URL}/realms/{KEYCLOAK_REALM}/protocol/openid-connect/userinfo"
OIDC_OP_DISCOVERY_DOCUMENT_URL = f"{KEYCLOAK_URL}/realms/{KEYCLOAK_REALM}/.well-known/openid-configuration"
AUTHENTICATION_BACKENDS = (
    "oauth2_authcodeflow.auth.AuthenticationBackend",
)

OIDC_RP_SCOPES = ["openid", "email", "profile"]
OIDC_MIDDLEWARE_NO_AUTH_URL_PATTERNS = ["/health", "/favicon.ico", "api/user/"]
OIDC_RP_USE_PKCE = False
OIDC_MIDDLEWARE_LOGIN_REQUIRED_REDIRECT = True
CORS_ALLOW_CREDENTIALS = True
CORS_ALLOWED_ORIGINS = [
    "http://localhost:3000",
    "http://localhost:8000",
    "*"
    # add any other hosts your front-end uses
]


JWT_SECRET_KEY = SECRET_KEY
JWT_ALGORITHM = 'HS256'
JWT_ACCESS_TOKEN_LIFETIME = timedelta(minutes=5)  # Short-lived
JWT_REFRESH_GRACE_PERIOD = timedelta(minutes=10)  # Grace period for refresh
JWT_EXPIRATION_HOURS = 2


REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "rest_framework.authentication.SessionAuthentication",
        # other authentication classes, if needed
    ],
    "DEFAULT_RENDERER_CLASSES": ["rest_framework.renderers.JSONRenderer"],
    "TEST_REQUEST_DEFAULT_FORMAT": "json",
    # FIXME: maybe use this at some point (for giving individual access rights):
    # "DEFAULT_PERMISSION_CLASSES": ['rest_framework_api_key.permissions.HasAPIKey' | 'rest_framework.permissions.IsAuthenticated']
}
API_KEY_CUSTOM_HEADER = "HTTP_API_KEY"

if os.getenv("SENDGRID_API_KEY", "envvar_not_existing") != "envvar_not_existing":
    SENDGRID_API_KEY = os.environ["SENDGRID_API_KEY"]
else:
    print("SENDGRID_API_KEY not found in environmental variables.")
    SENDGRID_API_KEY = "improperlyConfigured"

EMAIL_BACKEND = "sendgrid_backend.SendgridBackend"
SENDGRID_SANDBOX_MODE_IN_DEBUG = False
SENDGRID_ECHO_TO_STDOUT = True

# ##################################################
# ################## END AUTH ######################
# ##################################################

CORS_ORIGIN_WHITELIST = [
    "https://" + os.getenv("DOMAIN_HOSTED", "localhost:3000"),
    "http://localhost:3000",
    "http://127.0.0.1:3000",
    "http://localhost:8000",
    "*",
]

LANGUAGE_CODE = "en-us"

# TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

# TODO: does this fix the "Unauthorized: /api/model_tree/"-issue which occurs after some time??
TIME_ZONE = "Europe/Berlin"

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.0/howto/static-files/

STATIC_URL = "/static-django/"

STATIC_ROOT = "/static"

if os.getenv("STORAGE_TYPE") == "LEGACY" or not os.getenv("STORAGE_TYPE"):
    if os.getenv("KUBERNETES_ENGINE", "NONE") == "NONE":
        MEDIA_ROOT = os.path.join(NEW_BASE_DIR, f"{repo_name}/")
        MEDIA_URL = os.path.join(NEW_BASE_DIR, f"{repo_name}/")
        USER_REPORT_ROOT = os.path.join(NEW_BASE_DIR, f"{repo_name}/")
    else:
        # this section is not in effect when a cloud storage option is used.
        # Check DjangoProcessAdminGeneric/gcsUtils.py or
        # DjangoProcessAdminGeneric/sharepoint/SharePointCloudStorageUtils.py for replacement.
        MEDIA_ROOT = "/app/storage/uploads/"
        MEDIA_URL = "/app/storage/uploads/"

        USER_REPORT_ROOT = "/app/storage/reports/"

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        # Console stays plain
        "default": {
            "format": "{levelname} {asctime} {message}",
            "style": "{",
        },
        # Markdown formatter for WS – header in Markdown, message as-is
        "ws_md": {
            "format": ("\n{message}"),
            "style": "{",
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "default",
        },
        "ws": {
            "()": "lex.audit_logging.handlers.websocket_handler.WebSocketHandler",
            "level": "DEBUG",
            "formatter": "ws_md",
        },
    },
    "loggers": {
        "lex.calclog": {
            "handlers": ["ws", "console"],
            "level": "DEBUG",
            "propagate": False,
        },
    },
    "root": {
        "handlers": ["console"],
        "level": os.getenv("LOG_LEVEL", "DEBUG"),
    },
}

INSTANCE_CONTROLLER_BASE_URL = os.getenv(
    "INSTANCE_CONTROLLER_BASE_URL",
    "https://hazem.excellence-cloud.dev",
)

KEYCLOAK_SETUP_CALLBACK_URL = os.getenv(
    "KEYCLOAK_SETUP_CALLBACK_URL",
    "http://127.0.0.1:8000/api/bootstrap/keycloak/",
)


DATA_UPLOAD_MAX_MEMORY_SIZE = None

# Simple History Configuration
# Set to False to disable automatic exclusion of Django built-in models
SIMPLE_HISTORY_AUTO_EXCLUDE_BUILTIN = True

# Additional models to exclude from history tracking (beyond the built-in exclusions)
# Add model names (lowercase) that you specifically don't want to track
SIMPLE_HISTORY_EXCLUDED_MODELS = [
    # Example: 'temporarydata', 'cachedresult'
]



from urllib.parse import urlencode
from django.conf import settings


def provider_logout(request):
    """
    Build a Keycloak end-session URL that:
    1) passes the last ID token as a hint, and
    2) tells Keycloak where to send the user next.
    """
    base = settings.OIDC_OP_LOGOUT_ENDPOINT
    id_token = request.session.get("oidc_id_token")

    if not id_token:
        # fall back to simply clearing Django session
        return base

    # Keycloak (19+) expects:
    #  - id_token_hint
    #  - post_logout_redirect_uri
    params = {
        "id_token_hint": id_token,
        "post_logout_redirect_uri": settings.LOGOUT_REDIRECT_URL,
    }

    return f"{base}?{urlencode(params)}"



from lex.process_admin.views.model_relation_views import (
    ModelStructureObtainView,
    ModelStylingObtainView,
    Overview,
    ProcessStructure,
)

__all__ = [
    'ModelStructureObtainView',
    'ModelStylingObtainView',
    'Overview',
    'ProcessStructure',
]



def get_user_name(request):
    if "JIRA" in request.headers:
        return "Created by Jira"
    return f"{request.auth['name']} ({request.auth['sub']})"


def get_user_email(request):
    if "JIRA" in request.headers:
        return "jira@mail.com"
    return request.auth['email']



#!/usr/bin/env python3
import os
from pathlib import Path

from lex.tools.project_root import find_project_root  # shared utility

def generate_pycharm_configs(project_root=None):
    # Resolve against caller’s execution directory by default
    start = project_root or os.getcwd()
    project_root = os.path.abspath(find_project_root(start))

    runconfigs_dir = os.path.join(project_root, ".run")
    os.makedirs(runconfigs_dir, exist_ok=True)

    project_name = os.path.basename(project_root)
    env_file_path = os.path.join(project_root, ".env")
    env_files_option = (
        f'<option name="ENV_FILES" value="{env_file_path}" />'
        if os.path.exists(env_file_path) else
        '<option name="ENV_FILES" value="" />'
    )

    configs = {
        "Init.run.xml": {"name": "Init", "parameters": "Init"},
        "Start.run.xml": {
            "name": "Start",
            "parameters": "start --reload --loop asyncio lex_app.asgi:application",
        },
        "Make_migrations.run.xml": {"name": "Make migrations", "parameters": "makemigrations"},
        "Migrate.run.xml": {"name": "Migrate", "parameters": "migrate"},
        "Streamlit.run.xml": {"name": "Streamlit", "parameters": "streamlit run streamlit_app.py"},
        "Create_DB.run.xml": {
            "name": "Create DB",
            "parameters": "test lex.lex_app.logging.create_db.create_db --keepdb",
        },
        "Flush_DB.run.xml": {"name": "Flush DB", "parameters": "flush"},
    }

    print(f"Generating PyCharm run configurations in: {runconfigs_dir}")
    print(f"Project name: {project_name}")
    print(f"Project root: {project_root}")

    for filename, config in configs.items():
        content = f'''<component name="ProjectRunConfigurationManager">
  <configuration default="false" name="{config['name']}" type="PythonConfigurationType" factoryName="Python">
    <module name="{project_name}" />
    {env_files_option}
    <option name="INTERPRETER_OPTIONS" value="" />
    <option name="PARENT_ENVS" value="true" />
    <envs>
      <env name="PYTHONUNBUFFERED" value="1" />
    </envs>
    <option name="SDK_HOME" value="" />
    <option name="WORKING_DIRECTORY" value="{project_root}" />
    <option name="IS_MODULE_SDK" value="true" />
    <option name="ADD_CONTENT_ROOTS" value="true" />
    <option name="ADD_SOURCE_ROOTS" value="true" />
    <EXTENSION ID="PythonCoverageRunConfigurationExtension" runner="coverage.py" />
    <option name="SCRIPT_NAME" value="lex" />
    <option name="PARAMETERS" value="{config['parameters']}" />
    <option name="SHOW_COMMAND_LINE" value="false" />
    <option name="EMULATE_TERMINAL" value="false" />
    <option name="MODULE_MODE" value="true" />
    <option name="REDIRECT_INPUT" value="false" />
    <option name="INPUT_FILE" value="" />
    <method v="2" />
  </configuration>
</component>'''
        path = os.path.join(runconfigs_dir, filename)
        with open(path, "w", encoding="utf-8") as f:
            f.write(content)
        print(f"✓ Generated: {filename}")

    print("\nPyCharm run configurations generated successfully!")
    if os.path.exists(env_file_path):
        print(f"✓ Configurations will use .env file: {env_file_path}")
    else:
        print(f"⚠ No .env file found at {env_file_path}")
        print("  Create one if you need environment variables for your project.")

if __name__ == "__main__":
    import argparse, sys
    parser = argparse.ArgumentParser(description="Generate PyCharm run configurations for lex-app projects")
    parser.add_argument("-p", "--project-root", help="Project root directory (default: execution directory)")
    args = parser.parse_args()
    try:
        generate_pycharm_configs(args.project_root)
    except Exception as e:
        print(f"Error generating configurations: {e}")
        sys.exit(1)



"""
Configuration for django-simple-history to exclude Django built-in models
and other system models from history tracking.
"""

# Django built-in models that should NOT have history tracking
DJANGO_BUILTIN_MODELS = {
    # Django core models
    'contenttype',
    'permission',
    'group',
    'user',  # Only if you don't want User history - remove if you do
    'session',
    'logentry',
    
    # Django migrations
    'migration',
    
    # Django admin models
    'logentry',
    
    # Django auth models (optional - remove if you want history for these)
    'group_permissions',
    'user_groups',
    'user_user_permissions',
    
    # Django sites framework
    'site',
    
    # Django flatpages
    'flatpage',
    'flatpage_sites',
    
    # Django redirects
    'redirect',
    
    # Django cache framework
    'cache',
    
    # Django sessions
    'session',
    
    # OAuth2 authcodeflow models (if you don't want history for these)
    'blacklistedtoken',
    
    # REST framework models
    'apikey',
    
    # Celery models (if using django-celery-beat or similar)
    'crontabschedule',
    'intervalschedule',
    'periodictask',
    'solarschedule',
    'clockedschedule',
    'periodictasks',
    
    # Django channels (if using)
    'channellayer',
    
    # Simple history's own models
    'historicalrecord',
    
    # Add any other system models you want to exclude
}

# Third-party app models that should NOT have history tracking
THIRD_PARTY_EXCLUDED_MODELS = {
    # Add models from third-party packages that you don't want to track
    'corsheaders',
    'rest_framework',
    'oauth2_authcodeflow',
    'channels',
    'celery',
    'django_db_views',
}

# Your app-specific models that should NOT have history tracking
# This will be populated from Django settings
APP_EXCLUDED_MODELS = set()

def get_excluded_models():
    """
    Get the complete set of model names that should be excluded from history tracking.
    
    Returns:
        set: Set of lowercase model names to exclude from history tracking
    """
    from django.conf import settings
    
    excluded = set()
    
    # Add Django built-in models if auto-exclusion is enabled
    if getattr(settings, 'SIMPLE_HISTORY_AUTO_EXCLUDE_BUILTIN', True):
        excluded.update(DJANGO_BUILTIN_MODELS)
        excluded.update(THIRD_PARTY_EXCLUDED_MODELS)
    
    # Add app-specific excluded models from settings
    app_excluded = getattr(settings, 'SIMPLE_HISTORY_EXCLUDED_MODELS', [])
    excluded.update(set(model.lower() for model in app_excluded))
    
    # Add the original APP_EXCLUDED_MODELS
    excluded.update(APP_EXCLUDED_MODELS)
    
    return excluded

def should_track_model(model_class):
    """
    Determine if a model should have history tracking enabled.
    
    Args:
        model_class: Django model class to check
        
    Returns:
        bool: True if the model should have history tracking, False otherwise
    """
    from django.conf import settings
    
    model_name = model_class.__name__.lower()
    app_label = model_class._meta.app_label.lower()
    
    # CRITICAL: Never track Historical models themselves (prevents HistoricalHistorical* models)
    if model_class.__name__.startswith('Historical'):
        return False
    
    # Check if model name is in excluded list
    if model_name in get_excluded_models():
        return False
    
    # Check if auto-exclusion is enabled and it's from Django's built-in apps
    if getattr(settings, 'SIMPLE_HISTORY_AUTO_EXCLUDE_BUILTIN', True):
        django_apps = {
            'admin', 'auth', 'contenttypes', 'sessions', 'messages', 
            'staticfiles', 'sites', 'flatpages', 'redirects'
        }
        if app_label in django_apps:
            return False
        
        # Check if it's from excluded third-party apps
        excluded_apps = {
            'oauth2_authcodeflow', 'rest_framework', 'corsheaders', 
            'channels', 'celery', 'django_db_views', 'simple_history'
        }
        if app_label in excluded_apps:
            return False
    
    # Check if model is abstract
    if model_class._meta.abstract:
        return False
    
    # Check if model already has history (avoid double registration)
    if hasattr(model_class, 'history'):
        return False
    
    return True

def get_model_exclusion_reason(model_class):
    """
    Get the reason why a model is excluded from history tracking.
    Useful for debugging and logging.
    
    Args:
        model_class: Django model class to check
        
    Returns:
        str: Reason for exclusion, or None if model should be tracked
    """
    model_name = model_class.__name__.lower()
    app_label = model_class._meta.app_label.lower()
    
    # Check for Historical models first (most important check)
    if model_class.__name__.startswith('Historical'):
        return "Historical model (simple_history generated)"
    
    if model_name in DJANGO_BUILTIN_MODELS:
        return f"Django built-in model: {model_name}"
    
    if model_name in THIRD_PARTY_EXCLUDED_MODELS:
        return f"Third-party excluded model: {model_name}"
    
    if model_name in APP_EXCLUDED_MODELS:
        return f"App-specific excluded model: {model_name}"
    
    django_apps = {
        'admin', 'auth', 'contenttypes', 'sessions', 'messages', 
        'staticfiles', 'sites', 'flatpages', 'redirects'
    }
    if app_label in django_apps:
        return f"Django built-in app: {app_label}"
    
    excluded_apps = {
        'oauth2_authcodeflow', 'rest_framework', 'corsheaders', 
        'channels', 'celery', 'django_db_views', 'simple_history'
    }
    if app_label in excluded_apps:
        return f"Excluded third-party app: {app_label}"
    
    if model_class._meta.abstract:
        return "Abstract model"
    
    if hasattr(model_class, 'history'):
        return "Already has history tracking"
    
    return None  # Should be tracked